# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.aws.elb`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""

import logging
import math

from boto.exception import BotoServerError
from tornado import concurrent
from tornado import gen

from kingpin import utils
from kingpin.actors import exceptions
from kingpin.actors.aws import base
from kingpin.actors.utils import dry
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'


# This executor is used by the tornado.concurrent.run_on_executor()
# decorator. We would like this to be a class variable so its shared
# across RightScale objects, but we see testing IO errors when we
# do this.
EXECUTOR = concurrent.futures.ThreadPoolExecutor(10)


class CertNotFound(exceptions.UnrecoverableActorFailure):

    """Raised when an ELB is not found"""


# Helper function
def p2f(string):
    """Convert percentage string into float.

    Converts string like '78.9%' into 0.789
    """
    return float(string.strip('%')) / 100


class ELBBaseActor(base.AWSBaseActor):

    """Base class for ELB actors."""

    all_options = {
        'name': (str, REQUIRED, 'Name of the ELB'),
        'count': ((int, str), REQUIRED,
                  'Specific count, or percentage of instances to wait for.'),
        'region': (str, REQUIRED, 'AWS region (or zone) name, like us-west-2')
    }


class WaitUntilHealthy(ELBBaseActor):

    """Wait indefinitely until a specified ELB is considered "healthy".

    This actor will loop infinitely until a healthy threshold of the ELB is
    met.  The threshold can be reached when the ``count`` as specified in the
    options is less than or equal to the number of InService instances in the
    ELB.

    Another situation is for ``count`` to be a string specifying a percentage
    (see examples). In this case the percent of InService instances has to be
    greater than the ``count`` percentage.

    **Options**

    :name:
      The name of the ELB to operate on

    :count:
      Number, or percentage of InService instance to consider this ELB healthy

    :region:
      AWS region (or zone) name, such as us-east-1 or us-west-2

    **Examples**

    .. code-block:: json

       { "actor": "aws.elb.WaitUntilHealthy",
         "desc": "Wait until production-frontend has 16 hosts",
         "options": {
           "name": "production-frontend",
           "count": 16,
           "region": "us-west-2"
         }
       }

    .. code-block:: json

       { "actor": "aws.elb.WaitUntilHealthy",
         "desc": "Wait until production-frontend has 85% of hosts in-service",
         "options": {
           "name": "production-frontend",
           "count": "85%",
           "region": "us-west-2"
         }
       }

    **Dry Mode**

    This actor performs the finding of the ELB as well as calculating its
    health at all times. The only difference in dry mode is that it will not
    re-count the instances if the ELB is not healthy. A log message will be
    printed indicating that the run is dry, and the actor will exit with
    success.
    """

    desc = "Waiting until {name} is healthy ({count} in-service)"

    def _get_expected_count(self, count, total_count):
        """Calculate the expected count for a given percentage.

        Either returns the passed count if it's an integer, or
        calculates the count given an expected percentage.

        Args:
            count: Minimum count (int) or percentage (int) of hosts that must
                   be healthy.
            total_count: The total number of instances in the ELB.

        Returns:
            Number of instances required to be 'healthy'
        """

        if '%' in str(count):
            expected_count = math.ceil(total_count * p2f(count))
        else:
            expected_count = int(count)

        return expected_count

    @gen.coroutine
    def _is_healthy(self, elb, count):
        """Check if there are `count` InService instances for a given elb.

        Args:
            count: integer, or string with % in it.
                   for more information read _get_expected_count()

        Returns:
            Boolean whether or not the ELB is healthy enough.
        """
        name = elb.name

        self.log.debug('Counting ELB InService instances for : %s' % name)

        # Get all instances for this ELB
        instance_list = yield self.api_call(elb.get_instance_health)
        total_count = len(instance_list)

        self.log.debug('All instances: %s' % instance_list)
        in_service_count = [
            i.state for i in instance_list].count('InService')

        expected_count = self._get_expected_count(count, total_count)

        healthy = (in_service_count >= expected_count)
        self.log.debug('ELB "%s" healthy state: %s' % (elb.name, healthy))

        raise gen.Return(healthy)

    @gen.coroutine
    def _execute(self):
        """Executes an actor and yields the results when its finished.

        raises: gen.Return(True)
        """

        elb = yield self._find_elb(name=self.option('name'))

        repeating_log = utils.create_repeating_log(
            self.log.info,
            'Still waiting for %s to become healthy' % self.option('name'),
            seconds=30)
        while True:
            healthy = yield self._is_healthy(elb, count=self.option('count'))

            if healthy is True:
                self.log.info('ELB is healthy.')
                break

            # In dry mode, fake it
            if self._dry:
                self.log.info('Pretending that ELB is healthy.')
                break

            # Not healthy :( continue looping
            self.log.debug('Retrying in 3 seconds.')
            yield utils.tornado_sleep(3)

        utils.clear_repeating_log(repeating_log)

        raise gen.Return()


class SetCert(ELBBaseActor):

    """Find a server cert in IAM and use it for a specified ELB.

    **Options**

    :region:
      (str) AWS region (or zone) name, like us-west-2

    :name:
      (str) Name of the ELB

    :cert_name:
      (str) Unique IAM certificate name, or ARN

    :port:
      (int) Port associated with the cert.
      (default: 443)

    **Example**

    .. code-block:: json

       { "actor": "aws.elb.SetCert",
         "desc": "Run SetCert",
         "options": {
           "cert_name": "new-cert",
           "name": "some-elb",
           "region": "us-west-2"
         }
       }

    **Dry run**

    Will check that ELB and Cert names are existent, and will also check that
    the credentials provided for AWS have access to the new cert for ssl.
    """

    all_options = {
        'name': (str, REQUIRED, 'Name of the ELB'),
        'port': (int, 443, 'Port associated with the cert'),
        'region': (str, REQUIRED, 'AWS region (or zone) name, like us-west-2'),
        'cert_name': (str, REQUIRED, 'Unique IAM certificate name, or ARN'),
    }

    @gen.coroutine
    def _check_access(self, elb):
        """Perform a dummy operation to check credential accesss.

        Intended to be used in a dry run, this method attempts to perform an
        invalid set_listener call and monitors the output of the error. If the
        error is anything other than AccessDenied then the provided credentials
        are sufficient and we do nothing.

        Args:
            elb: boto LoadBalancer object.
        """
        try:
            # A blank ARN value should have code 'CertificateNotFound'
            # We're only checking if credentials have sufficient access
            yield self.api_call(
                elb.set_listener_SSL_certificate,
                self.option('port'),
                '')
        except BotoServerError as e:
            if e.error_code == 'AccessDenied':
                raise exceptions.InvalidCredentials(e)

    @gen.coroutine
    def _get_cert_arn(self, name):
        """Return a server_certificate ARN.

        Searches for a certificate object and returns the "ARN" value.

        Args:
            name: certificate name

        Raises:
            CertNotFound - if the name doesn't match an existing cert.

        Returns:
            string: the ARN value of the certificate
        """

        self.log.debug('Searching for cert "%s"...' % name)
        try:
            cert = yield self.api_call(
                self.iam_conn.get_server_certificate, name)
        except BotoServerError as e:
            raise CertNotFound(
                'Could not find cert %s. Reason: %s' % (name, e))

        # Get the ARN of this cert
        arn = cert['get_server_certificate_response'].get(
            'get_server_certificate_result').get(
            'server_certificate').get(
            'server_certificate_metadata').get('arn')

        raise gen.Return(arn)

    @gen.coroutine
    @dry('Would instruct {elb} to use cert: {arn}')
    def _use_cert(self, elb, arn):
        """Assign an ssl cert to a given ELB.

        Args:
            elb: boto elb object.
            arn: ARN for server certificate to use.
        """

        self.log.info('Setting ELB "%s" to use cert arn: %s' % (elb, arn))
        try:
            yield self.api_call(
                elb.set_listener_SSL_certificate, self.option('port'), arn)
        except BotoServerError as e:
            raise exceptions.RecoverableActorFailure(
                'Applying new SSL cert to %s failed: %s' % (elb, e))

    def _compare_certs(self, elb, new_arn):
        """Check if a given ELB is using a provided ARN for its certificate.

        Args:
            elb: boto elb object.
            new_arn: ARN for server certificate to use.

        Returns:
            boolean: used cert is same as the provided one.
        """

        ssl = [lis for lis in elb.listeners
               if lis[0] == self.option('port')][0]

        arn = ssl[4]

        return arn == new_arn

    @gen.coroutine
    def _execute(self):
        """Find ELB, and a Cert, then apply it."""
        elb = yield self._find_elb(self.option('name'))
        cert_arn = yield self._get_cert_arn(self.option('cert_name'))

        same_cert = self._compare_certs(elb, cert_arn)

        if same_cert:
            self.log.warning('ELB %s is already using this cert.' % elb)
            raise gen.Return()

        if self._dry:
            yield self._check_access(elb)

        yield self._use_cert(elb=elb, arn=cert_arn)


class RegisterInstance(base.AWSBaseActor):

    """Add an EC2 instance to a load balancer or target group.

    **Options**

    :elb:
      (str) Name of the ELB or the Target Group ARN

    :instances:
      (str, list) Instance id, or list of ids. Default "self" id.

    :region:
      (str) AWS region (or zone) name, like us-west-2

    :enable_zones:
      (bool) add all available AZ to the elb. Default: True

    **Example**

    .. code-block:: json

       { "actor": "aws.elb.RegisterInstance",
         "desc": "Run RegisterInstance",
         "options": {
           "elb": "prod-loadbalancer",
           "instances": "i-123456",
           "region": "us-east-1",
         }
       }

    .. code-block:: yaml

       ---
       actor: aws.elb.RegisterInstance
       desc: Run RegisterInstance
       options:
         elb: prod-loadbalancer
         instances: i-123456
         region: us-east-1

    **Dry run**

    Will find the specified ELB, but not take any actions regarding instances.
    """

    all_options = {
        'elb': (str, REQUIRED, 'Name of the ELB'),
        'region': (str, REQUIRED, 'AWS region (or zone) name, like us-west-2'),
        'instances': ((str, list), None, (
            'Instance id, or list of ids. If no value is specified then '
            'the instance id of the executing machine is used.')),
        'enable_zones': ((str, bool), True, 'Enable all zones for this ELB.')
    }

    @gen.coroutine
    @dry('Would add {instances} to {elb}')
    def _add(self, elb, instances):
        """Invoke elb.register_instances

        Args:
            elb: boto Loadbalancer object
            instances: list of instance ids.
        """
        yield self.api_call(elb.register_instances, instances)

    @gen.coroutine
    @dry('Would ensure {elb} is a member of all AZs')
    def _check_elb_zones(self, elb):
        """Ensure that `elb` has all available zones."""
        zones = yield self.api_call(self.ec2_conn.get_all_zones)
        zone_names = {z.name for z in zones}

        enabled_zones = set(elb.availability_zones)

        if not zone_names.issubset(enabled_zones):
            self.log.warning('ELB "%s" is missing some AZ.' % elb.name)
            self.log.info('Enabling all zones: %s' % zone_names)
            yield self.api_call(elb.enable_zones, zone_names)

    @gen.coroutine
    def _execute(self):
        elb = yield self._find_elb(self.option('elb'))
        instances = self.option('instances')

        if not instances:
            self.log.debug('No instance provided. Using current instance id.')
            iid = yield self._get_meta_data('instance-id')
            instances = [iid]
            self.log.debug('Instances is: %s' % instances)

        if type(instances) is not list:
            instances = [instances]

        self.log.info(('Adding the following instances to elb: '
                       '%s' % ', '.join(instances)))
        yield self._add(elb=elb, instances=instances)

        if self.str2bool(self.option('enable_zones')):
            yield self._check_elb_zones(elb=elb)


class DeregisterInstance(base.AWSBaseActor):

    """Remove EC2 instance(s) from an ELB.

    **Options**

    :elb:
      (str) Name of the ELB. Optionally this may also be a `*`.

    :instances:
      (str, list) Instance id, or list of ids

    :region:
      (str) AWS region (or zone) name, like us-west-2

    :wait_on_draining:
      (bool) Whether or not to wait for connection draining

    **Example**

    .. code-block:: json

       { "actor": "aws.elb.DeregisterInstance",
         "desc": "Run DeregisterInstance",
         "options": {
           "elb": "my-webserver-elb",
           "instances": "i-abcdeft",
           "region": "us-west-2"
         }
       }

    .. code-block:: yaml

       ---
       actor: aws.elb.DeregisterInstance
       desc: Run DeregisterInstance
       options:
         elb: prod-loadbalancer
         instances: i-123456
         region: us-east-1

    Extremely simple way to remove the local instance running this code from
    all ELBs its been joined to:

    .. code-block:: json

       { "actor": "aws.elb.DeregisterInstance",
         "desc": "Run DeregisterInstance",
         "options": {
           "elb": "*",
           "region": "us-west-2"
         }
       }

    **Dry run**

    Will find the ELB but not take any actions regarding the instances.
    """

    all_options = {
        'elb': (str, REQUIRED, 'Name of the ELB'),
        'region': (str, REQUIRED, 'AWS region (or zone) name, like us-west-2'),
        'instances': ((str, list), None, (
            'Instance id, or list of ids. If no value is specified then '
            'the instance id of the executing machine is used.')),
        'wait_on_draining': ((str, bool), True, (
            'Whether or not to wait for the ELB to drain connections '
            'before returning from the actor.'))
    }

    @gen.coroutine
    @dry('Would remove instances from {elb}: {instances}')
    def _remove(self, elb, instances):
        """Invoke elb.deregister_instances

        Args:
            elb: boto Loadbalancer object
            instances: list of instance ids.
        """
        self.log.info(('Removing instances from %s: %s'
                      % (elb, ', '.join(instances))))

        yield self.api_call(elb.deregister_instances, instances)
        yield self._wait_on_draining(elb)

    @gen.coroutine
    def _wait_on_draining(self, elb):
        """Waits for the ELB Connection Draining to occur.

        ELB Connection Draining is a configured-setting on the ELB that will
        continue to allow existing connections to be handeled before finally
        cutting them off at the timeout. This method will detect if connection
        draining is enabled, and optionally "sleep" for that time period before
        returning from the actor.

        Args:
            elb: boto Loadbalancer object
        """
        if not self.str2bool(self.option('wait_on_draining')):
            self.log.warning('Not waiting for connections to drain!')

        attrs = yield self.api_call(elb.get_attributes)
        if attrs.connection_draining.enabled:
            timeout = attrs.connection_draining.timeout

            self.log.info('Connection Draining Enabled, waiting %s(s)'
                          % timeout)
            yield utils.tornado_sleep(timeout)

    @gen.coroutine
    def _find_instance_elbs(self, instances):
        """Finds all ELBs that Instances are members of.

        Searches through all of the ELBs in a particular region and looks for
        which ones have any of the instances supplied in them. Creates a list
        of the ELBs, and returns the entire list.

        Args:
            instances: A list of Instance IDs

        Returns:
            a list of LoadBalancer objects
        """
        all_elbs = yield self.api_call(self.elb_conn.get_all_load_balancers)
        elbs_with_members = []

        for instance in instances:
            elbs = [lb for lb in all_elbs
                    if instance in [i.id for i in lb.instances]]
            self.log.debug('%s is a member of %s' % (instance, elbs))
            elbs_with_members.extend(elbs)

        raise gen.Return(elbs_with_members)

    @gen.coroutine
    def _execute(self):
        instances = self.option('instances')

        if not instances:
            self.log.debug('No instance provided. Using current instance id.')
            iid = yield self._get_meta_data('instance-id')
            instances = [iid]
            self.log.debug('Instances is: %s' % instances)

        if type(instances) is not list:
            instances = [instances]

        if self.option('elb') == '*':
            elbs = yield self._find_instance_elbs(instances)
        else:
            elb = yield self._find_elb(self.option('elb'))
            elbs = [elb]

        tasks = []
        for elb in elbs:
            tasks.append(self._remove(elb=elb, instances=instances))

        yield tasks



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.aws.cloudformation`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""

import logging
import json
import uuid

from botocore.exceptions import ClientError
from tornado import concurrent
from tornado import gen
from tornado import ioloop

from kingpin import utils
from kingpin.actors import exceptions
from kingpin.actors.utils import dry
from kingpin.actors.aws import base
from kingpin.constants import SchemaCompareBase, StringCompareBase
from kingpin.constants import REQUIRED, STATE

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


# This executor is used by the tornado.concurrent.run_on_executor()
# decorator. We would like this to be a class variable so its shared
# across RightScale objects, but we see testing IO errors when we
# do this.
EXECUTOR = concurrent.futures.ThreadPoolExecutor(10)


class CloudFormationError(exceptions.RecoverableActorFailure):

    """Raised on any generic CloudFormation error."""


class StackFailed(exceptions.RecoverableActorFailure):

    """Raised any time a Stack fails to be created or updated."""


class InvalidTemplate(exceptions.UnrecoverableActorFailure):

    """An invalid CloudFormation template was supplied."""


class StackAlreadyExists(exceptions.RecoverableActorFailure):

    """The requested CloudFormation stack already exists."""


class StackNotFound(exceptions.RecoverableActorFailure):

    """The requested CloudFormation stack does not exist."""


class ParametersConfig(SchemaCompareBase):

    """Validates the Parameters option.

    A valid `parameters` option is a dictionary with simple Key/Value pairs of
    strings. No nested dicts, arrays or other objects.
    """

    SCHEMA = {
        'type': ['object', 'null'],
        'uniqueItems': True,
        'patternProperties': {
            '.*': {
                'type': 'string'
            }

        }
    }

    valid = '{ "key", "value", "key2", "value2" }'


class CapabilitiesConfig(SchemaCompareBase):

    """Validates the Capabilities option"""

    SCHEMA = {
        'type': ['array', 'null'],
        'uniqueItems': True,
        'items': {
            'type': 'string',
            'enum': ['CAPABILITY_IAM',
                     'CAPABILITY_NAMED_IAM',
                     'CAPABILITY_AUTO_EXPAND']
        }
    }
    valid = '[ "CAPABILITY_IAM", "CAPABILITY_NAMED_IAM" ]'


class OnFailureConfig(StringCompareBase):

    """Validates the On Failure option.

    The `on_failure` option can take one of the following settings:
    `DO_NOTHING`, `ROLLBACK`, `DELETE`

    This option is applied at stack _creation_ time!
    """

    valid = ('DO_NOTHING', 'ROLLBACK', 'DELETE')


class TerminationProtectionConfig(StringCompareBase):

    """Validates the TerminationProtectionConfig option.

    The `enable_termination_protection` option can take one of the following
    settings:
    `'UNCHANGED'`, `False`, `True`

    `UNCHANGED` means on Create Stack it will default to False, however on
     Ensure Stack no changes will be applied.
    """

    valid = ('UNCHANGED', True, False)


# CloudFormation has over a dozen different 'stack states'... but for the
# purposes of these actors, we really only care about a few logical states.
# Here we map the raw states into logical states.
COMPLETE = (
    'CREATE_COMPLETE', 'UPDATE_COMPLETE', 'UPDATE_ROLLBACK_COMPLETE',
    'IMPORT_COMPLETE', 'IMPORT_ROLLBACK_COMPLETE')
DELETED = ('DELETE_COMPLETE', )
IN_PROGRESS = (
    'CREATE_PENDING', 'CREATE_IN_PROGRESS', 'DELETE_IN_PROGRESS',
    'EXECUTE_IN_PROGRESS', 'ROLLBACK_IN_PROGRESS',
    'UPDATE_COMPLETE_CLEANUP_IN_PROGRESS', 'UPDATE_IN_PROGRESS',
    'UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS',
    'UPDATE_ROLLBACK_IN_PROGRESS', 'IMPORT_IN_PROGRESS',
    'IMPORT_ROLLBACK_IN_PROGRESS')
FAILED = (
    'CREATE_FAILED', 'DELETE_FAILED', 'ROLLBACK_FAILED',
    'UPDATE_ROLLBACK_FAILED', 'ROLLBACK_COMPLETE',
    'IMPORT_ROLLBACK_FAILED')


class CloudFormationBaseActor(base.AWSBaseActor):

    """Base Actor for CloudFormation tasks"""

    # Get references to existing objects that are used by the
    # tornado.concurrent.run_on_executor() decorator.
    ioloop = ioloop.IOLoop.current()

    executor = EXECUTOR

    # Used mainly for unit testing..
    all_options = {
        'region': (str, REQUIRED, 'AWS region (or zone) name, like us-west-2')
    }

    def _discover_noecho_params(self, template_body):
        """Scans a CF template for NoEcho parameters.

        Searches through a CloudFormation stack template body for any
        parameters that are defined with the NoEcho flag. If there are any,
        returns a list of those parameter names.

        Args:
            template_body: (Str) CloudFormation Template Body

        Returns:
            A list of parameters that have NoEcho set to True
        """
        template = json.loads(template_body)
        stack_params = template.get('Parameters', {})
        noecho_params = [k for k in stack_params if
                         stack_params[k].get('NoEcho', False) is True]
        return noecho_params

    def _discover_default_params(self, template_body):
        """Scans a CF template for Default parameters.

        Searches through a CloudFormation stack template body for any
        parameters that are defined with the Default flag. If there are any,
        returns a dict of those parameter names mapped to their default values.

        Args:
            template_body: (Str) CloudFormation Template Body

        Returns:
            A dict of parameters with defaults mapped to their default values
        """
        template = json.loads(template_body)
        stack_params = template.get('Parameters', {})
        default_params = {
            k: stack_params[k]['Default'] for k in stack_params if
            stack_params[k].get('Default', None) is not None
        }
        return default_params

    def _get_template_body(self, template):
        """Reads in a local template file and returns the contents.

        If the template string supplied is a local file resource (has no
        URI prefix), then this method will return the contents of the file.
        Otherwise, returns None.

        Args:
            template: String with a reference to a template location.

        Returns:
            One tuple of:
              (Contents of template file, None)
              (None, URL of template)

        Raises:
            InvalidTemplate
        """
        if template is None:
            return None, None

        remote_types = ('http://', 'https://')

        if template.startswith(remote_types):
            return None, template

        try:
            return json.dumps(self._parse_policy_json(template)), None
        except exceptions.UnrecoverableActorFailure as e:
            raise InvalidTemplate(e)

    @gen.coroutine
    def _validate_template(self, body=None, url=None):
        """Validates the CloudFormation template.

        args:
          body: The body of the template
          url: A URL pointing to a template

        Raises:
            InvalidTemplate
            exceptions.InvalidCredentials
        """

        if body is not None:
            cfg = {'TemplateBody': body}
            self.log.info('Validating template with AWS...')
            try:
                yield self.api_call(self.cf3_conn.validate_template, **cfg)
            except ClientError as e:
                raise InvalidTemplate(e)

        if url is not None:
            cfg = {'TemplateURL': url}
            self.log.info('Validating template (%s) with AWS...' % url)
            try:
                yield self.api_call(self.cf3_conn.validate_template, **cfg)
            except ClientError as e:
                raise InvalidTemplate(e)

    def _create_parameters(self, parameters):
        """Converts a simple Key/Value dict into Amazon CF Parameters.

        The Boto3 interface requires that Parameters are passed in like this:

        .. code-block:: python
            Parameters=[
                { 'ParameterKey': 'string',
                  'ParameterValue': 'string',
                },
            ]

        This method takes a simple Dict of Key/Value pairs and converts it into
        the above format.

        Args:
            parameters: A dict of key/values

        Returns:
            A list like above
        """

        new_params = [
            {'ParameterKey': k,
             'ParameterValue': v}
            for k, v in list(parameters.items())]
        sorted_params = sorted(new_params, key=lambda k: k['ParameterKey'])
        return sorted_params

    @gen.coroutine
    def _get_stack(self, stack):
        """Returns a cloudformation.Stack object of the requested stack.

        If a "stack name" is supplied, Amazon returns only stacks that are
        "live" -- it does not return deleted stacks. If a "stack id" is used,
        Amazon will return the deleted stack as well.

        Args:
            stack: Stack name or stack ID

        Returns
            <Stack Dict> or <None>
        """
        try:
            stacks = yield self.api_call_with_queueing(
                self.cf3_conn.describe_stacks,
                queue_name='describe_stacks',
                StackName=stack)
        except ClientError as e:
            if 'does not exist' in str(e):
                raise gen.Return(None)

            raise CloudFormationError(e)

        raise gen.Return(stacks['Stacks'][0])

    @gen.coroutine
    def _get_stack_template(self, stack):
        """Returns the live policy attached to a CF Stack.

        args:
            stack: Stack name or stack ID
        """
        try:
            ret = yield self.api_call(self.cf3_conn.get_template,
                                      StackName=stack,
                                      TemplateStage='Original')
        except ClientError as e:
            raise CloudFormationError(e)

        raise gen.Return(ret['TemplateBody'])

    @gen.coroutine
    def _wait_until_state(self, stack_name, desired_states, sleep=15):
        """Indefinite loop until a stack has finished creating/deleting.

        Whether the stack has failed, suceeded or been rolled back... this
        method loops until the process has finished. If the final status is a
        failure (rollback/failed) then an exception is raised.

        Args:
            stack_name: The stack name or stack ID to watch
            desired_states: (tuple/list) States that indicate a successful
                            operation.
            sleep: (int) Time in seconds between stack status checks

        Raises:
            StackNotFound: If the stack doesn't exist.
        """
        while True:
            stack = yield self._get_stack(stack_name)

            if not stack:
                msg = 'Stack "%s" not found.' % self.option('name')
                raise StackNotFound(msg)

            # First, lets see if the stack is still in progress (either
            # creation, deletion, or rollback .. doesn't really matter)
            if stack['StackStatus'] in IN_PROGRESS:
                self.log.info('Stack state is %s, waiting %s(s)...' %
                              (stack['StackStatus'], sleep))
                yield utils.tornado_sleep(sleep)
                continue

            # If the stack is in the desired state, then return
            if stack['StackStatus'] in desired_states:
                self.log.debug('Found Stack state: %s' % stack['StackStatus'])
                raise gen.Return()

            # Lastly, if we get here, then something is very wrong and we got
            # some funky status back. Throw an exception.
            msg = 'Unexpected Stack state (StackStatus) received (%s): %s' % (
                stack['StackStatus'],
                stack.get('StackStatusReason',
                          'StackStatusReason not provided.'))
            raise StackFailed(msg)

    @gen.coroutine
    def _get_stack_events(self, stack):
        """Returns a list of human-readable CF Events.

        Searches for all of the Stack events for a given CF Stack and returns
        them in a human-readable list of strings.

        http://docs.aws.amazon.com/AWSCloudFormation/latest/
        APIReference/API_DescribeStackEvents.html

        args:
            stack: Stack ID or Stack name

        returns:
            [<list of human readable strings>]
        """
        try:
            raw = yield self.api_call(
                self.cf3_conn.describe_stack_events, StackName=stack)
        except ClientError:
            raise gen.Return([])

        # Reverse the list, and iterate through the data
        events = []
        for event in raw['StackEvents'][::-1]:
            # Not every event has a "reason" ... for those, we add a blank
            # reason value just to make string formatting easier below.
            if 'ResourceStatusReason' not in event:
                event['ResourceStatusReason'] = ''

            log_string_fmt = (
                '{ResourceType} {LogicalResourceId} '
                '({ResourceStatus}): {ResourceStatusReason}'
            )

            events.append(log_string_fmt.format(**event))

        raise gen.Return(events)

    @gen.coroutine
    @dry('Would have deleted stack {stack}')
    def _delete_stack(self, stack):
        """Executes the stack deletion."""

        exists = yield self._get_stack(stack)
        if not exists:
            raise StackNotFound('Stack does not exist!')

        self.log.info('Deleting stack')
        try:
            ret = yield self.api_call(
                self.cf3_conn.delete_stack, StackName=stack)
        except ClientError as e:
            raise CloudFormationError(str(e))

        req_id = ret['ResponseMetadata']['RequestId']
        self.log.info('Stack delete requested: %s' % req_id)

        # Now wait until the stack creation has finished
        try:
            yield self._wait_until_state(exists['StackId'], DELETED)
        except StackNotFound:
            # Pass here because a stack not found exception is totally
            # reasonable since we're deleting the stack. Sometimes Amazon
            # actually deletes the stack immediately, and othertimes it lists
            # the stack as a 'deleted' state, but we still get that state back.
            # Either case is fine.
            pass

    @gen.coroutine
    @dry('Would have created stack {stack}')
    def _create_stack(self, stack):
        """Executes the stack creation."""
        # Create the stack, and get its ID.
        self.log.info('Creating stack %s' % stack)

        cfg = dict()
        if self._template_body:
            cfg['TemplateBody'] = self._template_body
        else:
            cfg['TemplateURL'] = self._template_url

        if self.option('role_arn'):
            cfg['RoleARN'] = self.option('role_arn')

        enable_termination_protection = self.option(
            'enable_termination_protection')
        if enable_termination_protection == 'UNCHANGED':
            enable_termination_protection = False

        try:
            stack = yield self.api_call(
                self.cf3_conn.create_stack,
                StackName=stack,
                Parameters=self._parameters,
                OnFailure=self.option('on_failure'),
                TimeoutInMinutes=self.option('timeout_in_minutes'),
                Capabilities=self.option('capabilities'),
                EnableTerminationProtection=enable_termination_protection,
                **cfg)
        except ClientError as e:
            raise CloudFormationError(str(e))

        # Now wait until the stack creation has finished. If the creation
        # fails, get the logs from Amazon for the user.
        try:
            yield self._wait_until_state(stack['StackId'], COMPLETE)
        except StackFailed as e:
            events = yield self._get_stack_events(stack['StackId'])
            for e in events:
                self.log.error(e)
            msg = 'Stack creation failed: %s' % events
            raise StackFailed(msg)

        self.log.info('Stack created: %s' % stack['StackId'])

        raise gen.Return(stack['StackId'])


class Create(CloudFormationBaseActor):
    """Creates a CloudFormation stack.

    Creates a CloudFormation stack from scratch and waits until the stack is
    fully built before exiting the actor.

    **Options**

    :name:
      The name of the queue to create

    :capabilities:
      A list of CF capabilities to add to the stack.

    :on_failure:
     (:py:class:`OnFailureConfig`)

     One of the following strings: `DO_NOTHING`, `ROLLBACK`, `DELETE`

     Default: `DELETE`

    :parameters:
      A dictionary of key/value pairs used to fill in the parameters for the
      CloudFormation template.

    :region:
      AWS region (or zone) string, like 'us-west-2'.

    :role_arn:
      The Amazon IAM Role to use when executing the stack.

    :template:
      String of path to CloudFormation template. Can either be in the form of a
      local file path (ie, `./my_template.json`) or a URI (ie
      `https://my_site.com/cf.json`).

    :timeout_in_minutes:
      The amount of time that can pass before the stack status becomes
      CREATE_FAILED.

    :enable_termination_protection:
      Whether termination protection is enabled for the stack.

    **Examples**

    .. code-block:: json

       {
         "actor": "aws.cloudformation.Create",
         "desc": "Create production backend stack",
         "options": {
           "capabilities": [ "CAPABILITY_IAM" ],
           "name": "%CF_NAME%",
           "parameters": {
             "test_param": "%TEST_PARAM_NAME%",
           },
           "region": "us-west-1",
           "role_arn": "arn:aws:iam::123456789012:role/DeployRole",
           "template": "/examples/cloudformation_test.json",
           "timeout_in_minutes": 45,
           "enable_termination_protection": true,
         }
       }

    **Dry Mode**

    Validates the template, verifies that an existing stack with that name does
    not exist. Does not create the stack.
    """

    all_options = {
        'capabilities': (list, [],
                         'The list of capabilities that you want to allow '
                         'in the stack'),
        'on_failure': (OnFailureConfig, 'DELETE',
                       'Action to take if the stack fails to be created'),
        'name': (str, REQUIRED, 'Name of the stack'),
        'parameters': (ParametersConfig, {}, 'Parameters passed into the CF '
                                             'template execution'),
        'region': (str, REQUIRED, 'AWS region (or zone) name, like us-west-2'),
        'role_arn': (str, None,
                     'The Amazon IAM Role to use when executing the stack'),
        'template': (str, REQUIRED,
                     'Path to the AWS CloudFormation File. http(s)://, '
                     'file:///, absolute or relative file paths.'),
        'timeout_in_minutes': (int, 60,
                               'The amount of time that can pass before the '
                               'stack status becomes CREATE_FAILED'),
        'enable_termination_protection': (TerminationProtectionConfig,
                                          'UNCHANGED',
                                          'Whether termination protection is '
                                          'enabled for the stack.')
    }

    desc = "Creating CloudFormation Stack {name}"

    def __init__(self, *args, **kwargs):
        """Initialize our object variables."""
        super(Create, self).__init__(*args, **kwargs)

        # Convert our supplied parameters into a properly formatted list.
        self._parameters = self._create_parameters(self.option('parameters'))

        # Check if the supplied CF template is a local file. If it is, read it
        # into memory.
        self._template_body, self._template_url = self._get_template_body(
            self.option('template'))

    @gen.coroutine
    def _execute(self):
        stack_name = self.option('name')

        yield self._validate_template(
            self._template_body,
            self._template_url)

        # If a stack already exists, we cannot re-create it. Raise a
        # recoverable exception and let the end user decide whether this is bad
        # or not.
        exists = yield self._get_stack(stack_name)
        if exists:
            raise StackAlreadyExists('Stack %s already exists!' % stack_name)

        # If we're in dry mode, exit at this point. We can't do anything
        # further to validate that the creation process will work.
        if self._dry:
            self.log.info('Skipping CloudFormation Stack creation.')
            raise gen.Return()

        # Create the stack
        yield self._create_stack(stack=stack_name)

        raise gen.Return()


class Delete(CloudFormationBaseActor):
    """Deletes a CloudFormation stack

    **Options**

    :name:
      The name of the queue to create

    :region:
      AWS region (or zone) string, like 'us-west-2'

    **Examples**

    .. code-block:: json

       { "desc": "Delete production backend stack",
         "actor": "aws.cloudformation.Create",
         "options" {
           "region": "us-west-1",
           "name": "%CF_NAME%",
         }
       }

    **Dry Mode**

    Validates that the CF stack exists, but does not delete it.
    """

    all_options = {
        'name': (str, REQUIRED, 'Name of the stack'),
        'region': (str, REQUIRED, 'AWS region (or zone) name, like us-west-2')
    }

    desc = "Deleting CloudFormation Stack {name}"

    @gen.coroutine
    def _execute(self):
        stack_name = self.option('name')
        yield self._delete_stack(stack=stack_name)


class Stack(CloudFormationBaseActor):
    """Manages the state of a CloudFormation stack.

    This actor can manage the following aspects of a CloudFormation stack in
    Amazon:

      * Ensure that the Stack is present or absent.
      * Monitor and update the stack Template and Parameters as necessary.

    **Default Parameters**

    If your CF stack defines parameters with defaults, Kingpin will use the
    defaults unless the parameters are explicitly specified.

    **NoEcho Parameters**

    If your CF stack takes a Password as a parameter or any other value thats
    secret and you set `NoEcho: True` on that parameter, Kingpin will be unable
    to diff it and compare whether or not the desired setting matches whats in
    Amazon. A warning will be thrown, and the rest of the actor will continue
    to operate as normal.

    If any other difference triggers a Stack Update, the desired value for the
    parameter with `NoEcho: True` will be pushed in addition to all of the
    other stack parameters.

    **Options**

    :name:
      The name of the queue to create

    :state:
      (str) Present or Absent. Default: "present"

    :capabilities:
      (:py:class:`CapabilitiesConfig`, None)

      A list of CF capabilities to add to the stack.

    :disable_rollback:
      Set to True to disable rollback of the stack if creation failed.

    :on_failure:
     (:py:class:`OnFailureConfig`, None)

     One of the following strings: `DO_NOTHING`, `ROLLBACK`, `DELETE`

     Default: `DELETE`

    :parameters:
      (:py:class:`ParametersConfig`, None)

      A dictionary of key/value pairs used to fill in the parameters for the
      CloudFormation template.

    :region:
      AWS region (or zone) string, like 'us-west-2'.

    :role_arn:
      The Amazon IAM Role to use when executing the stack.

    :template:
      String of path to CloudFormation template. Can either be in the form of a
      local file path (ie, `./my_template.json`) or a URI (ie
      `https://my_site.com/cf.json`).

    :timeout_in_minutes:
      The amount of time that can pass before the stack status becomes
      CREATE_FAILED.

    :enable_termination_protection:
      Whether termination protection is enabled for the stack.

    **Examples**

    .. code-block:: json

       {
         "actor": "aws.cloudformation.Stack",
         "desc": "Manages the state of a CloudFormation stack",
         "options": {
           "capabilities": [ "CAPABILITY_IAM" ],
           "on_failure": "DELETE",
           "name": "%CF_NAME%",
           "parameters": {
             "test_param": "%TEST_PARAM_NAME%",
           },
           "region": "us-west-1",
           "role_arn": "arn:aws:iam::123456789012:role/DeployRole",
           "state": "present",
           "template": "/examples/cloudformation_test.json",
           "timeout_in_minutes": 45,
           "enable_termination_protection": true,
         }
       }

    **Dry Mode**

    Validates the template, verifies that an existing stack with that name does
    not exist. Does not create the stack.
    """

    all_options = {
        'name': (str, REQUIRED, 'Name of the stack'),
        'state': (STATE, 'present',
                  'Desired state of the bucket: present/absent'),
        'capabilities': (CapabilitiesConfig, [],
                         'The list of capabilities that you want to allow '
                         'in the stack'),
        'disable_rollback': (bool, False,
                             'Set to `True` to disable rollback of the stack '
                             'if stack creation failed.'),
        'on_failure': (OnFailureConfig, 'DELETE',
                       'Action to take if the stack fails to be created'),
        'parameters': (ParametersConfig, {}, 'Parameters passed into the CF '
                                             'template execution'),
        'region': (str, REQUIRED, 'AWS region (or zone) name, like us-west-2'),
        'role_arn': (str, None,
                     'The Amazon IAM Role to use when executing the stack'),
        'template': (str, REQUIRED,
                     'Path to the AWS CloudFormation File. http(s)://, '
                     'file:///, absolute or relative file paths.'),
        'timeout_in_minutes': (int, 60,
                               'The amount of time that can pass before the '
                               'stack status becomes CREATE_FAILED'),
        'enable_termination_protection': (TerminationProtectionConfig,
                                          'UNCHANGED',
                                          'Whether termination protection is '
                                          'enabled for the stack.')
    }

    desc = 'CloudFormation Stack {name}'

    def __init__(self, *args, **kwargs):
        """Initialize our object variables."""
        super(Stack, self).__init__(*args, **kwargs)

        # Check if the supplied CF template is a local file. If it is, read it
        # into memory.
        self._template_body, self._template_url = self._get_template_body(
            self.option('template'))

        # Find any Default parameters embedded in the stack.
        _default_params = self._discover_default_params(
            self._template_body)

        # Convert Default parameters and our supplied parameters into a
        # properly formatted list.
        # Defaults will be overridden by supplied parameters.
        self._parameters = self._create_parameters(
            dict(_default_params, **self.option('parameters')))

        # Discover whether or not there are any NoEcho parameters embedded in
        # the stack. If there are, record them locally and throw a warning to
        # the user about it.
        self._noecho_params = self._discover_noecho_params(self._template_body)
        for p in self._noecho_params:
            self.log.warning('Parameter "%s" has NoEcho set to True. '
                             'Will not use in parameter comparison.' % p)

    @gen.coroutine
    def _update_stack(self, stack):
        self.log.info('Verifying that stack is in desired state')

        # First, check that this stack isn't one that may have failed before
        # and there was attempted to be deleted but failed. If it is, we have a
        # fatal error and we must raise an exception.
        if stack['StackStatus'] == 'DELETE_FAILED':
            msg = 'Stack found in a deleted failed state: %s' % (
                stack['StackStatus'])
            raise StackFailed(msg)

        # Upon a stack creation, there are two states the stack can be left in
        # that are both un-fixable -- CREATE_FAILED and ROLLBACK_COMPLETE. In
        # both of these cases, the only possible option is to destroy the stack
        # and re-create it, you cannot fix a broken stack.
        if stack['StackStatus'] in ('CREATE_FAILED', 'ROLLBACK_COMPLETE'):
            self.log.warning(
                'Stack found in a failed state: %s' % stack['StackStatus'])
            yield self._delete_stack(stack=stack['StackId'])
            yield self._create_stack(stack=stack['StackName'])
            raise gen.Return()

        # Compare the live and new EnableTerminationProtection parameter and
        # update it if it is different.
        yield self._ensure_termination_protection(stack)

        # Pull down the live stack template and compare it to the one we have
        # locally.
        yield self._ensure_template(stack)

    @gen.coroutine
    def _ensure_template(self, stack):
        """Compares and updates the state of a CF Stack template

        Compares the current template body against the template body for the
        live running stack. If they're different. Triggers a Change Set
        creation and ultimately executes the change set.

        TODO: Support remote template_url comparison!

        args:
            stack: A Boto3 Stack object
        """
        needs_update = False

        # TODO: Implement this
        if self._template_url:
            self.log.warning('Cannot compare against remote template url')
            raise gen.Return()

        # Get the current template for the stack, and get our local template
        # body. Make sure they're in the same form (dict).
        existing = yield self._get_stack_template(stack['StackId'])
        new = json.loads(self._template_body)

        # Compare the two templates. If they differ at all, log it out for the
        # user and flip the needs_update bit.
        diff = utils.diff_dicts(existing, new)
        if diff:
            self.log.warning('Stack templates do not match.')
            for line in diff.split('\n'):
                self.log.info('Diff: %s' % line)

            # Plan to make a change set!
            needs_update = True

        # Get and compare the parameters we have vs the ones in CF. If they're
        # different, plan to do an update!
        if self._diff_params_safely(
                stack.get('Parameters', []),
                self._parameters):
            needs_update = True

        # If needs_update isn't set, then the templates are the same and we can
        # bail!
        if not needs_update:
            self.log.debug('Stack matches configuration, no changes necessary')
            raise gen.Return()

        # If we're here, the templates have diverged. Generate the change set,
        # log out the changes, and execute them.
        change_set_req = yield self._create_change_set(stack)
        change_set = yield self._wait_until_change_set_ready(
            change_set_req['Id'], 'Status', 'CREATE_COMPLETE')
        self._print_change_set(change_set)

        # Ok run the change set itself!
        try:
            yield self._execute_change_set(
                change_set_name=change_set_req['Id'])
        except (ClientError, StackFailed) as e:
            raise StackFailed(e)

        # In dry mode, delete our change set so we don't leave it around as
        # cruft. THis isn't necessary in the real run, because the changeset
        # cannot be deleted once its been applied.
        if self._dry:
            yield self.api_call(self.cf3_conn.delete_change_set,
                                ChangeSetName=change_set_req['Id'])

        self.log.info('Done updating template')

    def _diff_params_safely(self, remote, local):
        """Safely diffs the CloudFormation parameters.

        Does a comparison of the locally supplied parameters, and the remotely
        discovered (already set) CloudFormation parameters. When they are
        different, shows a clean diff and returns False.

        Takes into account NoEcho parameters which cannot be diff'd, so should
        not be included in the output (likely because they are passwords).

        Args:
            Remote: A list of objects, each having a ParameterKey and
            ParameterValue.
            Local: A list of objects, each having a ParameterKey and
            ParameterValue.

        Returns:
            Boolean
        """
        # If there are any NoEcho parameters, we can't diff them .. Amazon
        # returns them as *****'s and we're unable to compare them. Also, we
        # wouldn't want to print these out in our logs because they're almost
        # certainly passwords. Therefore, we should simply skip them in the
        # diff.
        for p in self._noecho_params:
            self.log.debug(
                'Removing "%s" from parameters before comparison.' % p)
            remote = [pair for pair in remote if pair['ParameterKey'] != p]
            local = [pair for pair in local if pair['ParameterKey'] != p]

        # Remove any resolved parameter values that were inserted by SSM
        # so that only supplied parameter values are compared.
        filtered_remote = []
        for param in remote:
            filtered_param = {}
            for k, v in param.items():
                if k != "ResolvedValue":
                    filtered_param[k] = v
            filtered_remote.append(filtered_param)

        remote = filtered_remote

        diff = utils.diff_dicts(remote, local)
        if diff:
            self.log.warning('Stack parameters do not match.')
            for line in diff.split('\n'):
                self.log.info('Diff: %s' % line)

            return True

        return False

    @gen.coroutine
    def _create_change_set(self, stack, uuid=uuid.uuid4().hex):
        """Generates a Change Set.

        Takes the current settings (template, capabilities, etc) and generates
        a Change Set against the live running stack. Returns back a Change Set
        Request dict, which can be used to poll for a real change set.

        args:
            stack: Boto3 Stack dict

        returns:
            Boto3 Change Set Request dict
        """
        change_opts = {
            'StackName': stack['StackId'],
            'Capabilities': self.option('capabilities'),
            'ChangeSetName': 'kingpin-%s' % uuid,
            'Parameters': self._parameters,
            'UsePreviousTemplate': False,
        }

        if self.option('role_arn'):
            change_opts['RoleARN'] = self.option('role_arn')

        if self._template_body:
            change_opts['TemplateBody'] = self._template_body
        else:
            change_opts['TemplateURL'] = self._template_url

        self.log.info('Generating a stack Change Set...')
        try:
            change_set_req = yield self.api_call(
                self.cf3_conn.create_change_set,
                **change_opts)
        except ClientError as e:
            raise CloudFormationError(e)

        raise gen.Return(change_set_req)

    @gen.coroutine
    def _wait_until_change_set_ready(self, change_set_name, status_key,
                                     desired_state, sleep=5):
        """Waits until a Change Set has hit the desired state.

        This loop waits until a Change Set has reached a desired state by
        comparing the value of the `status_key` with the `desired_state`. This
        allows the method to be used to check the status of the Change Set
        generation itself (status_key=Status) as well as the execution of the
        Change Set (status_key=ExecutionStatus).

        args:
            change_set_name: The Change Set Request Name
            status_key: The key within the Change Set that defines its status
            desired_state: A string of the desired state we're looking for
            sleep: Time to wait between checks in seconds

        returns:
            The final completed change set dictionary
        """
        self.log.info('Waiting for %s to reach %s' %
                      (change_set_name, desired_state))
        while True:
            try:
                change = yield self.api_call(
                    self.cf3_conn.describe_change_set,
                    ChangeSetName=change_set_name)
            except ClientError as e:
                # If we hit an intermittent error, lets just loop around and
                # try again.
                self.log.error('Error receiving Change Set state: %s' % e)
                yield utils.tornado_sleep(sleep)
                continue

            # The Stack State can be 'AVAILABLE', or an IN_PROGRESS string. In
            # either case, we loop and wait.
            if change[status_key] in (('AVAILABLE',) + IN_PROGRESS):
                self.log.info('Change Set state is %s, waiting %s(s)...' %
                              (change[status_key], sleep))
                yield utils.tornado_sleep(sleep)
                continue

            # If the stack is in the desired state, then return
            if change[status_key] == desired_state:
                self.log.debug('Change Set reached desired state: %s'
                               % change[status_key])
                raise gen.Return(change)

            # Lastly, if we get here, then something is very wrong and we got
            # some funky status back. Throw an exception.
            msg = 'Unexpected Change Set state (%s) received (%s): %s' % (
                status_key,
                change[status_key],
                change.get('StatusReason', 'StatusReason not provided.'))
            raise StackFailed(msg)

    def _print_change_set(self, change_set):
        """Logs out the changes a Change Set would make if executed.

        http://docs.aws.amazon.com/AWSCloudFormation/latest/
        APIReference/API_DescribeChangeSet.html

        args:
            change_set: Change Set Object
        """
        self.log.debug('Parsing change set: %s' % change_set)

        # Reverse the list, and iterate through the data
        for change in change_set['Changes']:
            resource = change['ResourceChange']

            if 'PhysicalResourceId' not in resource:
                resource['PhysicalResourceId'] = 'N/A'

            if 'Replacement' not in resource:
                resource['Replacement'] = False

            log_string_fmt = (
                'Change: '
                '{Action} {ResourceType} '
                '{LogicalResourceId}/{PhysicalResourceId} '
                '(Replacement? {Replacement})'
            )

            msg = log_string_fmt.format(**resource)
            self.log.warning(msg)

    @gen.coroutine
    @dry('Would have executed Change Set {change_set_name}')
    def _execute_change_set(self, change_set_name):
        """Executes the Change Set and waits for completion.

        Takes a supplied Change Set name and Stack Name, executes the change
        set, and waits for it to complete sucessfully.

        args:
            change_set_name: The Change Set Name/ARN
        """
        self.log.info('Executing change set %s' % change_set_name)
        try:
            yield self.api_call(self.cf3_conn.execute_change_set,
                                ChangeSetName=change_set_name)
        except ClientError as e:
            raise StackFailed(e)

        change_set = yield self._wait_until_change_set_ready(
            change_set_name, 'ExecutionStatus', 'EXECUTE_COMPLETE')
        yield self._wait_until_state(change_set['StackId'],
                                     (COMPLETE + FAILED + DELETED))

    @gen.coroutine
    def _ensure_termination_protection(self, stack):
        """Ensures that the EnableTerminationProtection is set to the desired
           setting (either True or False).

        Checks to to see if the actor is managing EnableTerminationProtection,
        and if it is, it updates EnableTerminationProtection if the defined
        value is different from the existing one.

        args:
            stack: Boto3 Stack dict
        """
        existing = stack['EnableTerminationProtection']
        new = self.option('enable_termination_protection')

        if new == 'UNCHANGED' or existing == new:
            raise gen.Return()

        yield self._update_termination_protection(stack, new)

    @gen.coroutine
    @dry('Would have updated EnableTerminationProtection')
    def _update_termination_protection(self, stack, new):
        """Updates the EnableTerminationProtection to the new setting.

        args:
            stack: Boto3 Stack dict
            new: boolean of updated value for EnableTerminationProtection
        """
        self.log.info('Updating EnableTerminationProtection to %s' % str(new))

        try:
            yield self.api_call(
                self.cf3_conn.update_termination_protection,
                StackName=stack['StackName'],
                EnableTerminationProtection=new)
        except ClientError as e:
            raise StackFailed(e)

    @gen.coroutine
    def _ensure_stack(self):
        state = self.option('state')
        stack_name = self.option('name')

        self.log.info('Ensuring that CF Stack %s is %s' %
                      (stack_name, state))

        # Figure out if the stack already exists or not. In this case, we
        # ignore DELETED stacks because they don't apply or block you from
        # creating a new stack.
        stack = yield self._get_stack(stack_name)

        # Before we figure out what to do, lets make sure the stack isn't in a
        # mutating state.
        if stack:
            yield self._wait_until_state(stack['StackId'],
                                         (COMPLETE + FAILED + DELETED))

        # Determine the current state of the stack vs the desired state
        if state == 'absent' and stack is None:
            self.log.debug('Stack does not exist')
        elif state == 'absent' and stack:
            yield self._delete_stack(stack=stack_name)
        elif state == 'present' and stack is None:
            stack = yield self._create_stack(stack=stack_name)
        elif state == 'present' and stack:
            stack = yield self._update_stack(stack)

        raise gen.Return(stack)

    @gen.coroutine
    def _execute(self):
        # Before we do anything, validate that the supplied template body or
        # url is valid. If its not, an exception is raised.
        yield self._validate_template(
            self._template_body,
            self._template_url)

        # This main method triggers the creation, deletion or update of the
        # stack as necessary.
        yield self._ensure_stack()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.aws.iam`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: User
   :noindex:

.. autoclass:: Group
   :noindex:

.. autoclass:: Role
   :noindex:

.. autoclass:: InstanceProfile
   :noindex:

.. autoclass:: UploadCert
   :noindex:

.. autoclass:: DeleteCert
   :noindex:
"""

import logging

# Bring in our sub class actors into the iam namespace
from kingpin.actors.aws.iam.certs import UploadCert, DeleteCert
from kingpin.actors.aws.iam.entities import User, Group, Role, InstanceProfile

log = logging.getLogger(__name__)

__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'

# Quiet down PyFlakes
User
Group
Role
InstanceProfile
UploadCert
DeleteCert



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.spotinst`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The Spotinst package allows you to create, manage and destroy Spotinst
ElastiGroups.

https://spotinst.atlassian.net/wiki/display/API/API+Semantics

**Environment Variables**

:SPOTINST_DEBUG:
  If set, then every single response body from Spotinst will be printed out in
  the debug logs for Kingpin. This can include credentials, and can be
  extremely verbose, so use with caution.

:SPOINST_TOKEN:
  SpotInst API Token generated at
  https://console.spotinst.com/#/settings/tokens

:SPOTINST_ACCOUNT_ID:
  SpotInst API Account ID - this is required unless you set the account_id
  parameter on each individual actor call.
  http://docs.spotinst.com/#page:api-semantic,header:header-organizations-with-a-single-account
"""

import base64
import copy
import logging
import os
import json

from tornado import gen
from tornado import httpclient

from tornado_rest_client import api

from kingpin import utils
from kingpin import exceptions as kingpin_exceptions
from kingpin.actors import base
from kingpin.actors import exceptions
from kingpin.actors.utils import dry
from kingpin.constants import REQUIRED
from kingpin.constants import SchemaCompareBase


log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'

DEBUG = os.getenv('SPOTINST_DEBUG', False)
TOKEN = os.getenv('SPOTINST_TOKEN', None)
ACCOUNT_ID = os.getenv('SPOTINST_ACCOUNT_ID', None)


class SpotinstAPI(api.RestConsumer):

    ENDPOINT = 'https://api.spotinst.io/'
    CONFIG = {
        'attrs': {
            'aws': {
                'new': True,
                'attrs': {
                    'ec2': {
                        'new': True,
                        'attrs': {
                            'list_groups': {
                                'new': True,
                                'path': 'aws/ec2/group?accountId=%account_id%',
                                'http_methods': {'get': {}}
                            },
                            'create_group': {
                                'new': True,
                                'path': 'aws/ec2/group?accountId=%account_id%',
                                'http_methods': {'post': {}}
                            },
                            'list_group': {
                                'path': 'aws/ec2/group/%id%?accountId=%account_id%',  # nopep8
                                'http_methods': {'get': {}}
                            },
                            'update_group': {
                                'path': 'aws/ec2/group/%id%?accountId=%account_id%',  # nopep8
                                'http_methods': {'put': {}}
                            },
                            'delete_group': {
                                'path': 'aws/ec2/group/%id%?accountId=%account_id%',  # nopep8
                                'http_methods': {'delete': {}}
                            },
                            'group_status': {
                                'path': 'aws/ec2/group/%id%/status?accountId=%account_id%',  # nopep8
                                'http_methods': {'get': {}}
                            },
                            'validate_group': {
                                'new': True,
                                'path': 'aws/ec2/group/validation?accountId=%account_id%',  # nopep8
                                'http_methods': {'post': {}}
                            },
                            'roll': {
                                'path': 'aws/ec2/group/%id%/roll?limit=50&accountId=%account_id%',  # nopep8
                                'http_methods': {'put': {}, 'get': {}}
                            },
                            'roll_status': {
                                'path': 'aws/ec2/group/%id%/roll/%roll_id%?accountId=%account_id%',  # nopep8
                                'http_methods': {'get': {}}
                            },
                        }
                    }
                }
            }
        }
    }


class SpotinstException(exceptions.RecoverableActorFailure):

    """Base SpotInst exception handler.

    This exception handler parses the Spotinst returned messages when an error
    is thrown. The error message comes back in the body in a JSON formatted
    blob. This exception handler will parse out the exception message, print it
    out in a semi-readable log form for the user, and then store it in the
    Exception body.

    See https://spotinst.atlassian.net/wiki/display/API/API+Semantics
    for more details.
    """

    def __init__(self, e):
        msg = self._parse(e)
        Exception.__init__(self, msg)
        self.exc = e

    def _parse(self, e):
        """Reads through a SpotInst error message body and parses it.

        This method looks for a proper error message(s) from Spotinst in the
        response body, parses them into something more humanly readable, and
        then logs them out. It also adds them to the exception message so that
        you get something beyond '400 Bad Request'.
        """
        log = logging.getLogger('%s.%s' % (self.__module__,
                                           self.__class__.__name__))
        try:
            error = json.loads(e.response.body)
        except AttributeError:
            return 'Unknown error: %s' % e

        msg_id = ('Request ID (%s) %s %s' % (error['request']['id'],
                                             error['request']['method'],
                                             error['request']['url']))
        log.error('Error on %s' % msg_id)

        if 'error' in error['response']:
            return 'Spotinst %s: %s' % (msg_id, error['response']['error'])

        if 'errors' in error['response']:
            msgs = []
            for err in error['response']['errors']:
                msg = '%s: %s' % (err['code'], err['message'])
                msgs.append(msg)
                log.error(msg)
            return 'Spotinst %s: %s' % (msg_id, ', '.join(msgs))

        # Fallback if we don't know what kind of error body this is
        error_str = ('Spotinst %s: %s' % (msg_id, error['response']))
        return error_str


class InvalidConfig(SpotinstException):

    """Thrown when an invalid request was supplied to Spotinst"""


class SpotinstRestClient(api.RestClient):

    EXCEPTIONS = {
        httpclient.HTTPError: {
            '400': InvalidConfig,
            '401': exceptions.InvalidCredentials,
            '403': exceptions.InvalidCredentials,
            '500': None,
            '502': None,
            '503': None,
            '504': None,

            # Represents a standard HTTP Timeout
            '599': None,

            '': exceptions.BadRequest,
        }
    }

    JSON_BODY = True

    TIMEOUT = 60


class ElastiGroupSchema(SchemaCompareBase):

    """Light validation against the Spotinst ElastiGroup schema.

    For full description of the JSON data format, please see:
    https://spotinst.atlassian.net/wiki/display/API/Create+Group#CreateGroup-JF

    This schema handles the following validation cases:

    * Only allow a single `SubnetID` for each `availabilityZone` object.
    * Disallow `t2|hc1` instance types for the `spot` instance section.
    * Ensure that the `scaling.up` and `scaling.down` arrays are either `null`
      or contain at least **1** record.
    """
    SCHEMA = {
        'type': 'object',
        'additionalProperties': True,
        'required': ['group'],
        'properties': {
            'group': {
                'type': 'object',
                'properties': {
                    'compute': {
                        'type': 'object',
                        'properties': {
                            'availabilityZones': {
                                'type': 'array',
                                'uniqueItems': True,
                                'items': {
                                    'type': 'object',
                                    'required': ['name', 'subnetId'],
                                    'additionalProperties': False,
                                    'properties': {
                                        'name': {'type': 'string'},
                                        'subnetId': {'type': 'string'}
                                    }
                                }
                            },
                            'instanceTypes': {
                                'type': 'object',
                                'properties': {
                                    'spot': {
                                        'type': 'array',
                                        'additionalItems': False,
                                        'items': {
                                            'type': 'string',
                                            'not': {
                                                'pattern': '^t2|hc1'
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                    'scaling': {
                        'type': ['object', 'null'],
                        'additionalProperties': False,
                        'properties': {
                            'up': {
                                'type': ['null', 'array'],
                                'minItems': 1
                            },
                            'down': {
                                'type': ['null', 'array'],
                                'minItems': 1
                            },
                        }
                    }
                }
            }
        }
    }


class SpotinstBase(base.EnsurableBaseActor):

    """Simple Spotinst Abstract Base Object"""

    def __init__(self, *args, **kwargs):
        super(SpotinstBase, self).__init__(*args, **kwargs)

        if not TOKEN:
            raise exceptions.InvalidCredentials(
                'Missing the "SPOTINST_TOKEN" environment variable.')

        if not DEBUG:
            logging.getLogger('tornado_rest_client.api').setLevel('INFO')

        # Figure out our account ID and set it.. Or this will end up falling
        # back to None if neither are set.
        account_id = self._options.get('account_id')
        if account_id is None:
            account_id = ACCOUNT_ID

        if account_id is None:
            raise exceptions.InvalidCredentials(
                'Missing SPOTINST_ACCOUNT_ID or account_id parameter')

        rest_client = SpotinstRestClient(
            headers={
                'Authorization': 'Bearer %s' % TOKEN,
                'Content-Type': 'application/json',
            })

        self._client = SpotinstAPI(client=rest_client, account_id=account_id)


class ElastiGroup(SpotinstBase):

    """Manages an ElastiGroup in Spotinst.

    `Spotinst ElastiGroups
    <https://spotinst.com/products/workload-management/elastigroup/>`_ act as
    smarter EC2 AutoScalingGroups that scale up and down leveraging Amazon Spot
    instances wherever possible. These ElastiGroups are completely configurable
    through a `JSON Blob
    <https://spotinst.atlassian.net/wiki/display/API/Create+Group>`_.

    For a fully functional example JSON config, see :download:`this one
    <../examples/test/spotinst.elastigroup/unittest.json>`. You can also write
    your files in YAML if you prefer -- Kingpin will handle the conversion.

    **UserData Startup Script**

    The Spotinst API wants the instances UserData script to be supplied as
    a Base64-encoded string -- which you can do if you wish. However, there is
    no need, as Kingpin will automatically convert your plain-text script into
    a Base64 blob for you behind the scenes.

    **Rolling out Group Changes**

    We will trigger the "roll group" API if the `roll_on_change` parameter is
    set to `True` after any change to an ElastiGroup. It is difficult to know
    which changes may or may not require a replacement of your existing hosts,
    so we leave this up to the user to decide on the behavior.

    **Known Limitations**

    * The Spotinst API does not allow you to change an ElastiGroup scaling
      'unit' (ie, CPU Count or Instance Count). You can also not change an
      ElastiGroup's basic platform (ie, VPC Linux vs Non VPC Linux). We warn
      about this on each change.

    **Options**

    :name:
      The desired name of the ElastiGroup. Note that this will override
      whatever value is inside your configuration JSON/YAML blob.

    :account_id:
      The SpotInst Account ID that the action is taking place in - this
      overrides the SPOTINST_ACCOUNT_ID environment variable (if its set).

    :config:
      Path to the ElastiGroup configuration blob (JSON or YAML) file.
      :ref:`token_replacement` can be used inside of your configuration files
      allowing environment variables to replace `%VAR%` strings.

      This file will be checked against a light-schema defined in
      :py:class:`ElastiGroupSchema` before any authentication is required. The
      file will be further validated against the Spotinst API during the DRY
      run, but this requires authentication.

    :tokens:
      A dict of key/value pairs that can be used to swap in variables into a
      common ElastiGroup template. These are added to (and override) the
      Environment variables that Kingpin already uses for variables swapping
      (as described in the :ref:`token_replacement` section.

    :roll_on_change:
      Whether or not to forcefully roll out changes to the ElastiGroup. If
      `True`, we will issue a 'roll call' to SpotInst and trigger all of the
      instances to be replaced. Defaults to `False`.

    :roll_batch_size:
      Indicates in percentage the amount of instances should be replaced in
      each batch. Defaults to `20`.

    :roll_grace_period:
      Indicates in seconds the timeout to wait until instance become healthy in
      the ELB. Defaults to `600`.

    :wait_on_create:
      If set to `True`, Kingpin will loop until the ElastiGroup has fully
      launched -- this only applies if the group is being created from scratch.
      On updates, see the `wait_on_roll` setting below.
      Defaults to `False`.

    :wait_on_roll:
      If set to `True`, Kingpin will loop until the rollout of any changes has
      completed. This can take a long time, depending on your rollout settings.
      Defaults to `False`.

    **Examples**

    .. code-block:: json

      { "actor": "spotinst.ElastiGroup",
        "options": {
          "name": "my-group",
          "config": "./group_config.json",
        }
      }

    **Dry Mode**

    Will discover the current state of the ElastiGroup (*present*, *absent*),
    and whether or not the current configuration is different than the desired
    configuration. Will also validate the desired configuration against the
    SpostInst API to give you a heads up about any potential failures up
    front.
    """

    all_options = {
        'name': (
            str, REQUIRED, 'Name of the ElastiGroup to manage'),
        'account_id': (
            str, None, 'SpotInst Account ID'),
        'config': (
            str, None, 'Name of the file with the ElastiGroup config'),
        'tokens': (
            dict, {}, ('A flat dictionary of Key/Value pairs that can be '
                       'swapped into the ElastiGroup template.')),
        'roll_on_change': (
            bool, False,
            ('Roll out new instances upon any config change.')),
        'roll_batch_size': (
            (str, int), 20,
            ('Indicates in percentage the amount of instances should be'
             'replaced in each batch.')),
        'roll_grace_period': (
            (str, int), 600,
            ('Indicates in seconds the timeout to wait until instance become'
             'healthy in the ELB.')),
        'wait_on_create': (
            bool, False, 'Wait for the ElastiGroup to startup and stabalize'),
        'wait_on_roll': (
            bool, False, 'Wait on any changes to roll out to the nodes'),
    }
    unmanaged_options = ['name', 'account_id', 'wait_on_roll',
                         'wait_on_create', 'roll_on_change', 'roll_batch_size',
                         'roll_grace_period', 'tokens']

    desc = 'ElastiGroup {name}'

    def __init__(self, *args, **kwargs):
        super(ElastiGroup, self).__init__(*args, **kwargs)

        # Quickly make sure that the roll_batch_size and roll_grace_period are
        # integers...
        for key in ('roll_batch_size', 'roll_grace_period'):
            try:
                self._options[key] = int(self._options[key])
            except ValueError:
                raise exceptions.InvalidOptions(
                    '%s (%s) must be an integer' % (key, self._options[key]))

        # Parse the user-supplied ElastiGroup config, swap in any tokens, etc.
        self._config = self._parse_group_config()

        # Filld in later by self._precache()
        self._group = None

    def _parse_group_config(self):
        """Parses the ElastiGroup config and replaces tokens.

        Reads through the supplied ElastiGroup configuration JSON blob (or
        YAML!), replaces any tokens that need replacement, and then sanity
        checks it against our schema.

        Note, contextual tokens (which are evaluated at run time, not
        compilation time) are not included here. Instead, those will be
        evaluated in the self._precache() method.
        """
        config = self.option('config')

        if config is None:
            return None

        self.log.debug('Parsing and validating %s' % config)

        # Join the init_tokens the class was instantiated with and the explicit
        # tokens that the user supplied.
        tokens = dict(self._init_tokens)
        tokens.update(self.option('tokens'))

        try:
            parsed = utils.convert_script_to_dict(
                script_file=config, tokens=tokens)
        except (kingpin_exceptions.InvalidScript, LookupError) as e:
            raise exceptions.InvalidOptions(
                'Error parsing %s: %s' % (config, e))

        # The userData portion of the body data needs to be Base64 encoded if
        # its not already. We will try to decode whatever is there, and if it
        # fails, we assume its raw text and we encode it.
        orig_data = (parsed['group']['compute']
                     ['launchSpecification']['userData'])
        new = base64.b64encode(orig_data.encode("utf-8"))
        parsed['group']['compute']['launchSpecification']['userData'] = new

        # Ensure that the name of the ElastiGroup in the config file matches
        # the name that was supplied to the actor -- or overwrite it.
        parsed['group']['name'] = self.option('name')

        # Now run the configuration through the schema validator
        ElastiGroupSchema.validate(parsed)

        return parsed

    @gen.coroutine
    def _precache(self):
        """Pre-populate a bunch of data.

        Searches for the list of ElastiGroups and stores the existing
        configuration for an ElastiGroup if it matches the name of the one
        we're managing here.

        Attempts light schema-validation of the desired ElastiGroup config to
        try to catch errors early in the Dry run.
        """
        # Check if the desired ElastiGroup already exists or not -- if it does,
        # store its configuration here for comparison purposes.
        self._group = yield self._get_group()

        # Validate the desired ElastiGroup configuration against the
        # schema-checker... light validation, but useful.
        yield self._validate_group()

        # Note - we don't manage the ElastiGroup target size. If the group
        # exists and has a target size set, we override the user-supplied
        # target number with the value returned to us by Spotinst.
        if self._group and 'capacity' in self._group['group']:
            target = self._group['group']['capacity']['target']
            self.log.info('Using the Spotinst supplied [capacity][target]'
                          ' value: %s' % target)
            self._config['group']['capacity']['target'] = target

    @gen.coroutine
    def _list_groups(self):
        """Returns a list of all ElastiGroups in your Spotinst acct.

        Returns:
            [List of JSON ElastiGroup objects]
        """
        raw = yield self._client.aws.ec2.list_groups.http_get()
        resp = raw.get('response', {})
        items = resp.get('items', [])
        raise gen.Return(items)

    @gen.coroutine
    def _get_group(self):
        """Finds and returns the existing ElastiGroup configuration.

        If the ElastiGroup exists, it returns the configuration for the group.
        If the group is missing, it returns None. Used by the self._precache()
        method to determine whether or not the desired ElastiGroup already
        exists or not, and what its configuration looks like.

        Returns:
          A dictionary with the ElastiGroup configuration returned by Spotinst
          or None if no matching group is found.

        Raises:
          exceptions.InvalidOptions: If too many groups are returned.
        """
        all_groups = yield self._list_groups()
        if not all_groups:
            raise gen.Return(None)

        matching = [
            group for group in all_groups
            if group['name'] == self.option('name')]

        if len(matching) > 1:
            raise exceptions.InvalidOptions(
                'Found more than one ElastiGroup with the name %s - '
                'this actor cannot manage multiple groups with the same'
                'name, you must use a unique name for each group.'
                % self.option('name'))

        if len(matching) < 1:
            self.log.debug('Did not find an existing ElastiGroup')
            raise gen.Return(None)

        match = matching[0]
        self.log.debug('Found ElastiGroup %s' % match['id'])
        raise gen.Return({'group': match})

    @gen.coroutine
    def _validate_group(self):
        """Basic Schema validation of the Elastigroup config.

        This endpoint is not documented, but it performs the most basic schema
        validation of the supplied ElastiGroup config. It cannot verify that
        instances will truly launch, but it can help catch obvious errors.

        It does require authentication, which is sad.

        Raises:
            SpotinstException: If any known Spotinst style error comes back.
        """
        yield self._client.aws.ec2.validate_group.http_post(
            group=self._config['group'])

    @gen.coroutine
    def _get_state(self):
        """Validates whether or not a matching ElastiGroup already exists.

        Depends on the self._precache() method having been called. If it has,
        then self._group should be populated if the group exists, or None if it
        doesn't.

        Returns:
            present: If the group exists
            absent: If not
        """
        if self._group:
            raise gen.Return('present')

        raise gen.Return('absent')

    @gen.coroutine
    def _set_state(self):
        """Creates or Deletes the ElastiGroup

        If the desired state is absent adn the group exists, we trigger a
        delete_group call. If the desired state is present and the group does
        not exist, we trigger a group create call. In any other situation, we
        do nothing because the desired and current states match.
        """
        if self.option('state') == 'absent' and self._group:
            yield self._delete_group(id=self._group['group']['id'])
        elif self.option('state') == 'present':
            yield self._create_group()

            # You'd think that we could store the returned group config from
            # Spotinst .. but it turns out that the data returned in the
            # create_group call above is not the same as what we've uploaded.
            # Instead, we have to re-call the self._precache() method to make
            # sure that we get an updated group config.
            # self._group = {'group': ret['response']['items'][0]}
            self._group = yield self._get_group()

            # Optionally, wait until the nodes have booted up before returning.
            if self.option('wait_on_create'):
                yield self._wait_until_stable()

    @gen.coroutine
    @dry('Would have created ElastiGroup')
    def _create_group(self):
        self.log.info('Creating ElastiGroup %s' % self.option('name'))
        yield self._client.aws.ec2.create_group.http_post(
            group=self._config['group'])

    @gen.coroutine
    @dry('Would have deleted ElastiGroup {id}')
    def _delete_group(self, id):
        self.log.info('Deleting ElastiGroup %s' % id)
        yield self._client.aws.ec2.delete_group(id=id).http_delete()

    @gen.coroutine
    def _get_group_status(self, id):
        self.log.debug('Getting ElastiGroup %s status...' % id)
        ret = yield self._client.aws.ec2.group_status(id=id).http_get()
        raise gen.Return(ret)

    @gen.coroutine
    def _get_config(self):
        """Not really used, but a stub for correctness"""
        raise gen.Return(self._group)

    @gen.coroutine
    @dry('Would have updated ElastiGroup config')
    def _set_config(self):
        group_id = self._group['group']['id']
        self.log.info('Updating ElastiGroup %s' % group_id)

        # There are certain fields that simply cannot be updated -- strip them
        # out. We have a warning up in the above _compare_config() section that
        # will tell the user about this in a dry run.
        if 'capacity' in self._config['group']:
            self.log.warning(
                'Note: Ignoring the group[capacity][unit] setting.')
            self._config['group']['capacity'].pop('unit', None)
        if 'compute' in self._config['group']:
            self.log.warning(
                'Note: Ignoring the group[compute][product] setting.')
            self._config['group']['compute'].pop('product', None)

        # Now do the update and capture the results. Once we have them, we'll
        # store the updated group configuration.
        ret = yield self._client.aws.ec2.update_group(id=group_id).http_put(
            group=self._config['group'])
        self._group = {'group': ret['response']['items'][0]}

        # If we're supposed to roll the group on any config changes, begin now
        if self.option('roll_on_change'):
            yield self._roll_group()

    @gen.coroutine
    def _compare_config(self):
        """Smart-ish comparison of Spotinst config to our own.

        This method is called by the EnsurableBaseClass to compare the desired
        (local) config with the existing (remote) config of the ElastiGroup.
        A simple == comparison will not work because there are additional
        fields returned by the Spotinst API (id, createdAt, updatedAt, and
        more) that will never be in the desired configuration object.

        This method makes copies of the configuration objects, strips out the
        fields that we cannot compare against, and then diffs. If a diff is
        detected, it logs out the diff for the end user, and then returns
        False.

        Returns:
            True: the configs match
            False: the configs do not match
        """
        # For the purpose of comparing the two configuration dicts, we need to
        # modify them (below).. so first lets copy them so we don't modify the
        # originals.
        new = copy.deepcopy(self._config)
        existing = copy.deepcopy(self._group)

        # If existing is none, then return .. there is no point in diffing the
        # config if the group doesn't exist! Note, this really only happens in
        # a dry run where we're creating the group because the group
        if existing is None:
            raise gen.Return(True)

        # Strip out some of the Spotinst generated and managed fields that
        # should never end up in either our new or existing configs.
        for field in ('id', 'createdAt', 'updatedAt', 'userData'):
            for g in (new, existing):
                g['group'].pop(field, None)

        # Decode both of the userData fields so we can actually see the
        # userdata differences.
        for config in (new, existing):
            config['group']['compute']['launchSpecification']['userData'] = (
                base64.b64decode(config['group']['compute']
                                 ['launchSpecification']['userData']))

        # We only allow a user to supply a single subnetId for each AZ (this is
        # handled by the ElastiGroupSchema). Spotinst returns back though both
        # the original setting, as well as a list of subnetIds. We purge that
        # from our comparison here.
        for az in existing['group']['compute']['availabilityZones']:
            az.pop('subnetIds', None)

        diff = utils.diff_dicts(existing, new)

        if diff:
            self.log.warning('Group configurations do not match')
            for line in diff.split('\n'):
                self.log.info('Diff: %s' % line)
            return False

        return True

    @gen.coroutine
    @dry('Would have rolled the ElastiGroup..')
    def _roll_group(self, delay=30):
        """Triggers an ElastiGroup rolling operation and waits for completion.

        Sends a signal to Spotinst to "roll" (replace) the nodes in the
        ElastiGroup based on the new configuration. This operation takes a
        while based on the `roll_batch_size` and `roll_grace_period` options.
        Depending on the `wait_on_roll` option, this method will wait until the
        roll has completed before returning.
        """
        group_id = self._group['group']['id']

        # You are not allowed to have two rolls happening at the same time --
        # so if there is already a roll in progress, we need to wait before we
        # issue another one. This is a requirement regardless of whether the
        # user has asked us to 'wait_on_roll' or not, because we'll get an
        # exception back from the API if we try to issue a roll call during an
        # existing roll operation.
        yield self._wait_until_roll_complete(delay)

        # Now, try to do the roll...
        self.log.info('Triggering an ElastiGroup roll')
        yield self._client.aws.ec2.roll(id=group_id).http_put(
            batchSizePercentage=self.option('roll_batch_size'),
            gracePeriod=self.option('roll_grace_period'))

        # Now, if the user wants us to wait, we will wait.
        if self.option('wait_on_roll'):
            yield self._wait_until_roll_complete(delay)

    @gen.coroutine
    @dry('Would have waited for ElastiGroup changes to become active')
    def _wait_until_roll_complete(self, delay):
        """Poll and wait until an ElastiGroup roll is complete.
        """
        group_id = self._group['group']['id']

        # Note: We do not use the repeating_log because we only call this API
        # every 30s or so. Rolling out group changes is almost guaranteed to be
        # a very slow process, so there is no need to make frequent API calls
        # to constantly check the status of the rollout. Instead, we make calls
        # infrequently and thus we are able to simply log out the status after
        # each call.
        self.log.info('Checking if any ElastiGroup rolls are in progress..')
        while True:
            response = yield self._client.aws.ec2.roll(id=group_id).http_get()

            in_progress = [r for r in response['response']['items']
                           if r['status'] != 'finished']

            if len(in_progress) < 1:
                break

            status = in_progress[0]['status']
            unit = in_progress[0]['progress']['unit']
            progress = in_progress[0]['progress']['value']

            self.log.info('Group roll is %s %s complete (%s)' % (progress,
                                                                 unit, status))

            yield gen.sleep(delay)

    @gen.coroutine
    @dry('Would have waited for all ElastiGroup nodes to launch')
    def _wait_until_stable(self, delay=3):
        """Poll and wait until an ElastiGroup has stabalized.

        Upon group creation, most of the instances will be in a "biding" state.
        This method watches the list of instances and waits until they are all
        in the 'fulfilled' state.
        """
        group_id = self._group['group']['id']

        # We use the repeating_log to let the user know we're still monitoring
        # things, while not  flooding them every time we make an API call. We
        # give them a message every 30s, but make an API call every 3 seconds
        # to check the status.
        repeating_log = utils.create_repeating_log(
            self.log.info,
            'Waiting for ElastiGroup to become stable',
            seconds=30)

        while True:
            response = yield self._get_group_status(group_id)

            # Find any nodes that are waiting for spot instance requests to be
            # fulfilled.
            pending = [i for i in response['response']['items']
                       if i['status'] == 'pending-evaluation']
            fulfilled = [i['instanceId'] for i in response['response']['items']
                         if i['status'] == 'fulfilled' and i['instanceId'] is
                         not None]

            if len(pending) < 1:
                self.log.info('All instance requests fulfilled: %s' %
                              ', '.join(fulfilled))
                break

            yield gen.sleep(delay)

        utils.clear_repeating_log(repeating_log)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

import jsonschema

from kingpin import exceptions

__author__ = 'Matt Wise <matt@nextdoor.com>'


ACTOR_SCHEMA = {
    'type': ['object'],
    'required': ['actor'],
    'additionalProperties': False,
    'properties': {
        'desc': {'type': 'string'},
        'actor': {'type': 'string'},
        'options': {
            # 'Options' are Actor specific. However, we validate some internal
            # option types here, if they are supplied.
            'type': 'object',

            # Since options are actor specific, ignore unexpected options.
            'additionalProperties': True,

            # Internally expected properties
            'properties': {
                # 'acts' are lists of actors that should be instantiated. Each
                # object should look like this actual schema (with a desc,
                # actor and option key)
                'acts': {
                    'type': 'array',

                    # This is a reference to 'self' ... in other words,
                    # this array can only contain other SCHEMA_1_0 type
                    # objets.
                    'items': {'$ref': '#'},
                },
            },
        },

        # Not required. In code, will default to False.
        'warn_on_failure': {'type': ['boolean', 'string']},

        # Not required. In code, will default to <actor>.default_timeout
        'timeout': {'type': ['string', 'integer', 'number']},

        # Optional conditional to indicate to skip this actor.
        'condition': {'type': ['boolean', 'string'], 'default': True},
    }
}

SCHEMA_1_0 = {
    'definitions': {
        'actor': ACTOR_SCHEMA
    },

    'anyOf': [
        {'$ref': '#/definitions/actor'},
        {'type': 'array', 'items': {'$ref': '#/definitions/actor'}}
    ]
}


def validate(config):
    """Validates the JSON against our schemas.

    TODO: Support multiple schema versions

    Args:
        config: Dictionary of parsed JSON

    Returns:
        None: if all is well

    Raises:
        Execption if something went wrong.
    """
    try:
        return jsonschema.validate(config, SCHEMA_1_0)
    except jsonschema.exceptions.ValidationError as e:
        raise exceptions.InvalidScript(e)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.aws.iam.entities`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""

import json
import os
import logging

from boto.exception import BotoServerError
from tornado import concurrent
from tornado import gen

from kingpin import utils
from kingpin.actors import exceptions
from kingpin.actors.aws.iam import base
from kingpin.constants import REQUIRED
from kingpin.constants import STATE

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


# This executor is used by the tornado.concurrent.run_on_executor()
# decorator. We would like this to be a class variable so its shared
# across RightScale objects, but we see testing IO errors when we
# do this.
EXECUTOR = concurrent.futures.ThreadPoolExecutor(10)

# The maximum number of items returned to us in a get_all_*/list_* api call.
# Defaults to 100, but setting to 1000 to reduce the number of API calls we
# make.
MAX_ITEMS = 1000


class EntityBaseActor(base.IAMBaseActor):

    """User/Group/Role Base Management Class

    Managing Users, Groups and Roles in Amazon IAM is nearly identical. This
    class abstracts that work, so that the actual User/Group/Role actors can be
    extremely simple and just handle the differences between each type of IAM
    entity.
    """

    all_options = {
        'name': (str, REQUIRED, 'The name of the user.'),
        'state': (STATE, 'present',
                  'Desired state of the User: present/absent'),
        'inline_policies': ((str, list), None,
                            'List of inline policy JSON files to apply.')
    }

    def __init__(self, *args, **kwargs):
        super(EntityBaseActor, self).__init__(*args, **kwargs)

        # These IAM Connection methods must be overridden in a subclass of this
        # actor. Each of these is a "generalized" name for the method in Boto
        # found at http://boto.cloudhackers.com/en/latest/ref/iam.html.
        #
        # This is a little confusing, but the idea is that these methods all
        # basically behave the same (are called the same way, return the same
        # type of data), so we should be able to generalize them into
        # variables.
        #
        # Once these are mapped to real IAM calls, then the methods in this
        # base class will work.

        # The "text name" of the entity type. This is either:
        #  user, group, role, instance_profile
        self.entity_name = 'base'

        self.create_entity = None
        self.delete_entity = None
        self.delete_entity_policy = None
        self.get_all_entities = None
        self.get_all_entity_policies = None
        self.get_entity_policy = None
        self.put_entity_policy = None

    def _generate_policy_name(self, policy):
        """Generates an Amazon-friendly Policy name from a filename.

        Amazon Inline IAM Policies have names -- and although allowing our
        users to enter their own name might be nice, its overkill in most
        cases. We'd rather just determine the name for them from the name of
        the policy definition file that they included in the JSON.

        http://docs.aws.amazon.com/IAM/latest/UserGuide/reference_iam-limits.html

        args:
            policy: The file name of the policy document

        returns:
            A string name to use as the policy name
        """
        # Get rid of the extension first
        name = os.path.splitext(policy)[0]

        # If theres a leading slash, strip it
        name = name.lstrip('/')

        # Replace slashes with dashes instead
        name = name.replace('/', '-')
        name = name.replace('\\', '-')

        # Strip out any non-allowed characters that made it through
        name = name.replace('*', '')
        name = name.replace('?', '')

        return name

    def _parse_inline_policies(self, policies):
        """Read, parse and store our inline policies.

        Any of the inline policies passed into this actor at init time are read
        in, parsed, turned into dicts and then stored in an object level
        dictionary for future use. This is done at __init__ time to make sure
        we catch any syntax errors as early as possible.

        args:
            policies: A string or list of strings that point to JSON files with
            IAM policies in them.
        """
        # If the inline_policies is None, then we bail and set
        # self.inline_policies to none.
        if policies is None:
            self.inline_policies = None
            return

        # Prepare to store our parsed inline policies in a hash of key/values
        # -- the key is the policy name (with no file ending) and the value is
        # the dict of the policy itself.
        self.inline_policies = {}

        # If a single policy was supplied (ie, maybe on a command line) then
        # turn it into a list.
        if isinstance(policies, str):
            policies = [policies]

        # Run through any supplied Inline IAM Policies and verify that they're
        # not corrupt very early on.
        for policy in policies:
            p_name = self._generate_policy_name(policy)
            self.inline_policies[p_name] = self._parse_policy_json(policy)

            self.log.debug('Parsed policy %s: %s' %
                           (p_name, self.inline_policies[p_name]))

    @gen.coroutine
    def _get_entity_policies(self, name):
        """Returns a dictionary of all the inline policies attached to a entity.

        args:
            name: The IAM Entity Name (Name/Group)

        returns:
            A dict of key/value pairs - key is the policy name, value is the
            dict-version of the policy document.
        """
        policies = {}

        # Get the list of inline policies attached to an entity. Note, not
        # all entities have a concept of inline policies. If
        # self.get_all_entity_policies is None, it returns a TypeError. We'll
        # catch that and silently move on.
        policy_names = []
        try:
            self.log.debug('Searching for any inline policies for %s' % name)
            ret = yield self.api_call(self.get_all_entity_policies, name)
            policy_names = (ret['list_%s_policies_response' % self.entity_name]
                               ['list_%s_policies_result' % self.entity_name]
                               ['policy_names'])
        except BotoServerError as e:
            if e.status == 404:
                # The user doesn't exist.. likely in a dry run. Return no
                # policies.
                policy_names = []
            else:
                raise exceptions.RecoverableActorFailure(
                    'An unexpected API error occurred: %s' % e)
        except TypeError:
            pass

        # Iterate through all of the named policies and fire off
        # get-requests, but don't yield on them yet.
        tasks = []
        for p_name in policy_names:
            tasks.append((p_name,
                          self.api_call(self.get_entity_policy, name, p_name)))

        # Now that we've fired off all the calls, we walk through each yielded
        # result, parse the returned policy, and append it to our policies
        # list. We also catch any raised exceptions here.
        for t in tasks:
            (p_name, p_task) = t
            try:
                raw = yield p_task
            except BotoServerError as e:
                raise exceptions.RecoverableActorFailure(
                    'An unexpected API error occurred downloading '
                    'policy %s: %s' % (p_name, e))

            # Convert the uuencoded doc string into a dict
            p_doc = self._policy_doc_to_dict((
                raw['get_%s_policy_response' % self.entity_name]
                   ['get_%s_policy_result' % self.entity_name]
                   ['policy_document']))

            # Store the converted document under the policy name key
            policies[p_name] = p_doc
            self.log.debug('Got policy %s/%s: %s' % (name, p_name, p_doc))

        raise gen.Return(policies)

    @gen.coroutine
    def _ensure_inline_policies(self, name):
        """Ensures that all of the inline IAM policies for a entity are managed

        This method has three stages.. first it ensures that any missing
        policies (as determined by the policy name) are applied to a entity.
        Second, it determines if any existing policies have changed locally and
        need to be updated in IAM. Finally it purges unmanaged policies that
        were applied to a entity out of band.

        args:
            name: The entity to manage
        """
        # Get the list of current entity policies first
        existing_policies = yield self._get_entity_policies(name)

        # First, push any policies that we have listed, but aren't in the
        # entity
        tasks = []
        for policy in (set(self.inline_policies.keys()) -
                       set(existing_policies.keys())):
            policy_doc = self.inline_policies[policy]
            tasks.append(self._put_entity_policy(name, policy, policy_doc))
        yield tasks

        # Do we have matching policies that we're managing here, and are
        # already attached to the entity profile? Lets make sure each one of
        # those matches the policy we have here, and update it if necessary.
        tasks = []
        for policy in (set(self.inline_policies.keys()) &
                       set(existing_policies.keys())):
            new = self.inline_policies[policy]
            exist = existing_policies[policy]
            diff = utils.diff_dicts(exist, new)
            if diff:
                self.log.info('Policy %s differs from Amazons:' % policy)
                for line in diff.split('\n'):
                    self.log.info('Diff: %s' % line)
                policy_doc = self.inline_policies[policy]
                tasks.append(self._put_entity_policy(name, policy, policy_doc))
        yield tasks

        # Purge any policies we found in AWS that were not listed in our actor
        tasks = []
        for policy in (set(existing_policies.keys()) -
                       set(self.inline_policies.keys())):
            tasks.append(self._delete_entity_policy(name, policy))
        yield tasks

    @gen.coroutine
    def _delete_entity_policy(self, name, policy_name):
        """Optionally pushes a policy to an IAM entity.

        args:
            name: The IAM Entity Name
            policy_name: The entity policy name
        """
        if self._dry:
            self.log.warning('Would delete policy %s from %s %s' %
                             (policy_name, self.entity_name, name))
            raise gen.Return()

        self.log.info('Deleting policy %s from %s %s' %
                      (policy_name, self.entity_name, name))
        try:
            ret = yield self.api_call(
                self.delete_entity_policy, name, policy_name)
            self.log.debug('Policy %s deleted: %s' % (policy_name, ret))
        except BotoServerError as e:
            if e.error_code != 404:
                raise exceptions.RecoverableActorFailure(
                    'An unexpected API error occurred: %s' % e)

    @gen.coroutine
    def _put_entity_policy(self, name, policy_name, policy_doc):
        """Optionally pushes a policy to an IAM Entity.

        args:
            name: The IAM Entity Name
            policy_name: The entity policy name
            policy_doc: The ploicy document object itself
        """
        if self._dry:
            self.log.warning('Would push policy %s to %s %s' %
                             (policy_name, self.entity_name, name))
            raise gen.Return()

        self.log.info('Pushing policy %s to %s %s' %
                      (policy_name, self.entity_name, name))
        try:
            ret = yield self.api_call(
                self.put_entity_policy,
                name,
                policy_name,
                json.dumps(policy_doc))
            self.log.debug('Policy %s pushed: %s' % (policy_name, ret))
        except BotoServerError as e:
            raise exceptions.RecoverableActorFailure(
                'An unexpected API error occurred: %s' % e)

    @gen.coroutine
    def _get_entity(self, name):
        """Returns an IAM Entity JSON Blob.

        Searches for an IAM Entity and either returns None, or a JSON blob that
        describes the Entity.

        args:
            name: The IAM Entity Name
        """
        self.log.debug('Searching for %s %s' % (self.entity_name, name))

        # Get a list of all of the entities - return 100 results at a time, and
        # paginate the results.
        is_truncated = True
        marker = None
        while is_truncated:
            # Get the list back - if the marker has been set, then we pass it
            # in and we start from where the last results told us we should.
            try:
                response = yield self.api_call(
                    self.get_all_entities, max_items=MAX_ITEMS, marker=marker)
            except BotoServerError as e:
                raise exceptions.RecoverableActorFailure(
                    'An unexpected API error occurred: %s' % e)

            # Get the result object from the response...
            result = (
                response['list_%ss_response' % self.entity_name]
                        ['list_%ss_result' % self.entity_name])

            # If the results indicate they were truncated, they'll include
            # a 'marker'. Setting these two variables will cause this to
            # loop again, in the event that we don't find the response in
            # the first set of results.
            is_truncated = self.str2bool(result.get('is_truncated', False))
            marker = result.get('marker', None)

            # Check our result for the entity.. if its there, great.
            # Otherwise, we'll move on.
            entity = [entity for entity in result['%ss' % self.entity_name]
                      if entity['%s_name' % self.entity_name] == name]

            if len(entity) > 0:
                self.log.debug(
                    'Found %s %s' % (self.entity_name, entity[0]['arn']))
                raise gen.Return(entity[0])

        # If there aren't any entities, return None.
        raise gen.Return()

    @gen.coroutine
    def _ensure_entity(self, name, state):
        """Ensures a entity is either present or absent.

        Looks up the entities current state and then makes a decision about
        creating or deleting the entity. If the entity is already in the
        correct state, not changes are made.

        args:
            name: The IAM User Name
            state: 'present' or 'absent'
        """
        self.log.info('Ensuring that %s %s is %s' %
                      (self.entity_name, name, state))

        entity = yield self._get_entity(name)

        if entity and state == 'present':
            raise gen.Return()
        elif not entity and state == 'present':
            yield self._create_entity(name)
        elif entity and state == 'absent':
            yield self._delete_entity(name)
        elif not entity and state == 'absent':
            raise gen.Return()

    @gen.coroutine
    def _create_entity(self, name):
        """Creates an IAM Entity.

        If the entity exists, we just warn and move on.

        args:
            name: The IAM Entity Name
        """
        if self._dry:
            self.log.warning('Would create %s %s' % (self.entity_name, name))
            raise gen.Return()

        try:
            ret = yield self.api_call(
                self.create_entity, name)
        except BotoServerError as e:
            if e.status != 409:
                raise exceptions.RecoverableActorFailure(
                    'An unexpected API error occurred: %s' % e)
            self.log.warning(
                '%s %s already exists, skipping creation.' %
                (self.entity_name, name))
            raise gen.Return()

        arn = (ret['create_%s_response' % self.entity_name]
                  ['create_%s_result' % self.entity_name]
                  [self.entity_name]['arn'])
        self.log.info('%s %s created' % (self.entity_name, arn))

    @gen.coroutine
    def _delete_entity(self, name):
        """Deletes and IAM Entity.

        If the entity doesn't exist, we just warn and move on.

        args:
            name: The IAM Entity Name
        """
        if self._dry:
            self.log.warning('Would delete %s %s' % (self.entity_name, name))
            raise gen.Return()

        try:
            # Get the entities policies. They have to be deleted before we can
            # possibly move forward and delete the entity.
            existing_policies = yield self._get_entity_policies(name)
            tasks = []
            for policy in existing_policies:
                tasks.append(self._delete_entity_policy(name, policy))
            yield tasks

            # Now delete the entity
            yield self.api_call(self.delete_entity, name)
            self.log.info('%s %s deleted' % (self.entity_name, name))
        except BotoServerError as e:
            if e.status != 404:
                raise exceptions.RecoverableActorFailure(
                    'An unexpected API error occurred: %s' % e)
            self.log.warning('%s %s doesn\'t exist' % (self.entity_name, name))

    @gen.coroutine
    def _add_user_to_group(self, name, group):
        """Quick helper method to add a user to a group.

        args:
            name: user name
            group: group name
        """
        if self._dry:
            self.log.warning('Would have added %s to %s' % (name, group))
            raise gen.Return()

        try:
            self.log.info('Adding %s to %s' % (name, group))
            yield self.api_call(self.iam_conn.add_user_to_group, group, name)
        except BotoServerError as e:
            raise exceptions.RecoverableActorFailure(
                'An unexpected API error occurred: %s' % e)

    @gen.coroutine
    def _remove_user_from_group(self, name, group):
        """Quick helper method to remove a user from a group.

        args:
            name: user name
            group: group name
        """
        if self._dry:
            self.log.warning('Would have removed %s from %s' % (name, group))
            raise gen.Return()

        try:
            self.log.info('Removing %s from %s' % (name, group))
            yield self.api_call(self.iam_conn.remove_user_from_group,
                                group, name)
        except BotoServerError as e:
            raise exceptions.RecoverableActorFailure(
                'An unexpected API error occurred: %s' % e)


class User(EntityBaseActor):

    """Manages an IAM User.

    This actor manages the state of an Amazon IAM User.

    Currently we can:

      * Ensure is present or absent
      * Manage the inline policies for the user
      * Manage the groups the user is in

    **Options**

    :name:
      (str) Name of the User profile to manage

    :state:
      (str) Present or Absent. Default: "present"

    :groups:
      (str,array) A list of groups for the user to be a member of.
      Default: None

    :inline_policies:
      (str,array) A list of strings that point to JSON files to use as inline
      policies.
      Default: None

    **Example**

    .. code-block:: json

       { "actor": "aws.iam.User",
         "desc": "Ensure that Bob exists",
         "options": {
           "name": "bob",
           "state": "present",
           "groups": "my-test-group",
           "inline_policies": [
             "read-all-s3.json",
             "create-other-stuff.json"
           ]
         }
       }

    **Dry run**

    Will let you know if the user exists or not, and what changes it would make
    to the users policy and settings. Will also parse the inline policies
    supplied, make sure any tokens in the files are replaced, and that the
    files are valid JSON.
    """

    all_options = {
        'name': (str, REQUIRED, 'The name of the user.'),
        'state': (STATE, 'present',
                  'Desired state of the User: present/absent'),
        'groups': ((str, list), None, 'List of groups to add the user to.'),
        'inline_policies': ((str, list), None,
                            'List of inline policy JSON files to apply.')
    }

    desc = "IAM User {name}"

    def __init__(self, *args, **kwargs):
        super(User, self).__init__(*args, **kwargs)

        self.entity_name = 'user'
        self.create_entity = self.iam_conn.create_user
        self.delete_entity = self.iam_conn.delete_user
        self.delete_entity_policy = self.iam_conn.delete_user_policy
        self.get_all_entities = self.iam_conn.get_all_users
        self.get_all_entity_policies = self.iam_conn.get_all_user_policies
        self.get_entity_policy = self.iam_conn.get_user_policy
        self.put_entity_policy = self.iam_conn.put_user_policy

        # Parse the supplied inline policies
        self._parse_inline_policies(self.option('inline_policies'))

    @gen.coroutine
    def _ensure_groups(self, name, groups):
        """Ensure that this user is a member of specific groups.

        args:
            name: The user we're managing
            groups: The list (or single) of groups to join be members of
        """
        if isinstance(groups, str):
            groups = [groups]

        current_groups = set()
        try:
            res = yield self.api_call(self.iam_conn.get_groups_for_user, name)
            current_groups = {g['group_name'] for g in
                              res['list_groups_for_user_response']
                                 ['list_groups_for_user_result']
                                 ['groups']}
        except BotoServerError as e:
            # If the error is a 404, then the user doesn't exist and we can
            # assume that the mappings don't exist at all. We leave the
            # existin_mappings list alone. For any other error, raise.
            if e.status != 404:
                raise exceptions.RecoverableActorFailure(
                    'An unexpected API error occurred: %s' % e)

        # Find any groups that we're not already a member of, and add us
        tasks = []
        try:
            for new_group in set(groups) - current_groups:
                tasks.append(self._add_user_to_group(name, new_group))
        except StopIteration:
            pass

        yield tasks

        # Find any group memberships we didn't know about, and purge them
        tasks = []
        for bad_group in current_groups - set(groups):
            tasks.append(self._remove_user_from_group(name, bad_group))

        yield tasks

    @gen.coroutine
    def _execute(self):
        name = self.option('name')
        state = self.option('state')
        groups = self.option('groups')

        yield self._ensure_entity(name, state)
        if state == 'absent':
            raise gen.Return()

        if self.option('inline_policies') is not None:
            yield self._ensure_inline_policies(name)

        if groups is not None:
            yield self._ensure_groups(name, groups)

        raise gen.Return()


class Group(EntityBaseActor):

    """Manages an IAM Group.

    This actor manages the state of an Amazon IAM Group.

    Currently we can:

      * Ensure is present or absent
      * Manage the inline policies for the group
      * Purge (or not) all group members and delete the group

    **Options**

    :name:
      (str) Name of the Group profile to manage

    :force:
      (bool) Forcefully delete the group (explicitly purging all group
      memberships).
      Default: false

    :state:
      (str) Present or Absent. Default: "present"

    :inline_policies:
      (str,array) A list of strings that point to JSON files to use as inline
      policies. You can also pass in a single inline policy as a string.
      Default: None

    **Example**

    .. code-block:: json

       { "actor": "aws.iam.Group",
         "desc": "Ensure that devtools exists",
         "options": {
           "name": "devtools",
           "state": "present",
           "inline_policies": [
             "read-all-s3.json",
             "create-other-stuff.json"
           ]
         }
       }

    **Dry run**

    Will let you know if the group exists or not, and what changes it would
    make to the groups policy and settings. Will also parse the inline policies
    supplied, make sure any tokens in the files are replaced, and that the
    files are valid JSON.
    """

    all_options = {
        'name': (str, REQUIRED, 'The name of the group.'),
        'force': (bool, False, 'Forcefully delete the group.'),
        'state': (STATE, 'present',
                  'Desired state of the group: present/absent'),
        'inline_policies': ((str, list), None,
                            'List of inline policy JSON files to apply.')
    }

    desc = "IAM Group {name}"

    def __init__(self, *args, **kwargs):
        super(Group, self).__init__(*args, **kwargs)

        self.entity_name = 'group'
        self.create_entity = self.iam_conn.create_group
        self.delete_entity = self.iam_conn.delete_group
        self.delete_entity_policy = self.iam_conn.delete_group_policy
        self.get_all_entities = self.iam_conn.get_all_groups
        self.get_all_entity_policies = self.iam_conn.get_all_group_policies
        self.get_entity_policy = self.iam_conn.get_group_policy
        self.put_entity_policy = self.iam_conn.put_group_policy

        # Parse the supplied inline policies
        self._parse_inline_policies(self.option('inline_policies'))

    @gen.coroutine
    def _get_group_users(self, name):
        """Returns a list of users assigned to the group.

        args:
            name: the name of the group

        returns:
            a list of user name strings
        """
        users = []
        try:
            raw = yield self.api_call(self.iam_conn.get_group, name)
            users = [user['user_name'] for user in
                     raw['get_group_response']['get_group_result']['users']]
        except BotoServerError as e:
            if e.status != 404:
                raise exceptions.RecoverableActorFailure(
                    'An unexpected API error occurred: %s' % e)
        except KeyError:
            # No users!
            users = []

        raise gen.Return(users)

    @gen.coroutine
    def _purge_group_users(self, name, force):
        """Forcefully purge all users from the group.

        This is used only if the group has users, is being deleted, and the
        'purge' option was set.

        args:
          name: the group name
          force: boolean whether or not to actually force the removal
        """
        users = yield self._get_group_users(name)

        if not force and users:
            self.log.warning(('Will not be able to delete this group '
                              'without first removing all of its members. '
                              'Use the `force` option to purge all members.'))
            self.log.warning('Group members: %s' % ', '.join(users))

        if not force:
            raise gen.Return()

        tasks = []
        for user in users:
            tasks.append(self._remove_user_from_group(user, name))
        yield tasks

    @gen.coroutine
    def _execute(self):
        name = self.option('name')
        state = self.option('state')
        force = self.option('force')

        if state == 'absent':
            yield self._purge_group_users(name, force)

        yield self._ensure_entity(name, state)
        if state == 'absent':
            raise gen.Return()

        if self.option('inline_policies') is not None:
            yield self._ensure_inline_policies(name)

        raise gen.Return()


class Role(EntityBaseActor):

    """Manages an IAM Role.

    This actor manages the state of an Amazon IAM Role.

    Currently we can:

      * Ensure is present or absent
      * Manage the inline policies for the role
      * Manage the Assume Role Policy Document

    **Options**

    :name:
      (str) Name of the Role to manage

    :state:
      (str) Present or Absent. Default: "present"

    :inline_policies:
      (str,array) A list of strings that point to JSON files to use as inline
      policies. You can also pass in a single inline policy as a string.
      Default: None

    :assume_role_policy_document:
      (str) A string with an Amazon IAM Assume Role policy. Not providing this
      causes Kingpin to ignore the value, and Amazon defaults the role to an
      'EC2' style rule. Supplying the document will cause Kingpin to ensure the
      assume role policy is correct.
      Default: None

    **Example**

    .. code-block:: json

       { "actor": "aws.iam.Role",
         "desc": "Ensure that myapp exists",
         "options": {
           "name": "myapp",
           "state": "present",
           "inline_policies": [
             "read-all-s3.json",
             "create-other-stuff.json"
           ]
         }
       }

    **Dry run**

    Will let you know if the group exists or not, and what changes it would
    make to the groups policy and settings. Will also parse the inline policies
    supplied, make sure any tokens in the files are replaced, and that the
    files are valid JSON.
    """

    all_options = {
        'name': (str, REQUIRED, 'The name of the role.'),
        'state': (STATE, 'present',
                  'Desired state of the group: present/absent'),
        'inline_policies': ((str, list), None,
                            'List of inline policy JSON files to apply.'),
        'assume_role_policy_document': (str, None,
                                        ('The policy that grants an entity'
                                         'permission to assume the role'))
    }

    desc = "IAM Role {name}"

    def __init__(self, *args, **kwargs):
        super(Role, self).__init__(*args, **kwargs)

        self.entity_name = 'role'
        self.create_entity = self.iam_conn.create_role
        self.delete_entity = self.iam_conn.delete_role
        self.delete_entity_policy = self.iam_conn.delete_role_policy
        self.get_all_entities = self.iam_conn.list_roles
        self.get_all_entity_policies = self.iam_conn.list_role_policies
        self.get_entity_policy = self.iam_conn.get_role_policy
        self.put_entity_policy = self.iam_conn.put_role_policy

        # Pre-parse the supplied inline policies
        self._parse_inline_policies(self.option('inline_policies'))

        # Pre-parse the Assume Role Policy Document if it was supplied
        if self.option('assume_role_policy_document') is not None:
            self.assume_role_policy_doc = self._parse_policy_json(
                self.option('assume_role_policy_document'))

    @gen.coroutine
    def _ensure_assume_role_doc(self, name):
        """Ensures that the Assume Role Policy for a Role is up to date.

        Downloads the existing Assume Role Policy for a given Role, then
        compares it against our configured policy and optionally updates it if
        they differ.

        Args:
            name: The role we're workin with
        """
        # Get our existing role policy from the entity
        entity = yield self._get_entity(name)

        # If the entity doesn't exist, then we must be in a Dry run and the
        # role hasn't been created yet. Just bail silently.
        if not entity:
            raise gen.Return()

        # Parse the raw data into a dict we can compare
        exist = self._policy_doc_to_dict(entity['assume_role_policy_document'])
        new = self.assume_role_policy_doc

        # Now diff it against our desired policy. If no diff, then quietly
        # return.
        diff = utils.diff_dicts(exist, new)
        if not diff:
            self.log.debug('Assume Role Policy documents match')
            raise gen.Return()

        self.log.info('Assume Role Policy differs from Amazons:')
        for line in diff.split('\n'):
            self.log.info('Diff: %s' % line)

        if self._dry:
            self.log.warning('Would have updated the Assume Role Policy Doc')
            raise gen.Return()

        self.log.info('Updating the Assume Role Policy Document')
        yield self.api_call(
            self.iam_conn.update_assume_role_policy, name, json.dumps(new))

    @gen.coroutine
    def _execute(self):
        name = self.option('name')
        state = self.option('state')

        yield self._ensure_entity(name, state)
        if state == 'absent':
            raise gen.Return()

        if self.option('inline_policies') is not None:
            yield self._ensure_inline_policies(name)

        if self.option('assume_role_policy_document') is not None:
            yield self._ensure_assume_role_doc(name)

        raise gen.Return()


class InstanceProfile(EntityBaseActor):

    """Manages an IAM Instance Profile.

    This actor manages the state of an Amazon IAM Instance Profile.

    Currently we can:

      * Ensure is present or absent
      * Assign an IAM Role to the Instance Profile

    **Options**

    :name:
      (str) Name of the Role to manage

    :state:
      (str) Present or Absent. Default: "present"

    :role:
      (str) Name of an IAM Role to assign to the Instance Profile.
      Default: None

    **Example**

    .. code-block:: json

       { "actor": "aws.iam.InstanceProfile",
         "desc": "Ensure that my-ecs-servers exists",
         "options": {
           "name": "my-ecs-servers",
           "state": "present",
           "role": "some-iam-role",
         }
       }

    **Dry run**

    Will let you know if the profile exists or not, and what changes it would
    make to the profile.
    """

    all_options = {
        'name': (str, REQUIRED, 'The name of the instance profile.'),
        'state': (STATE, 'present',
                  'Desired state of the group: present/absent'),
        'role': (str, None, 'Name of an IAM Role to assign')
    }

    desc = "InstanceProfile {name}"

    def __init__(self, *args, **kwargs):
        super(InstanceProfile, self).__init__(*args, **kwargs)

        self.entity_name = 'instance_profile'
        self.create_entity = self.iam_conn.create_instance_profile
        self.delete_entity = self.iam_conn.delete_instance_profile
        self.get_all_entities = self.iam_conn.list_instance_profiles

    @gen.coroutine
    def _add_role(self, name, role):
        """Adds a role to an Instance Profile.

        args:
            name: The name of the Instance Profile we're managing
            role: The name of the role to assign to the profile
        """
        if self._dry:
            self.log.warning('Would add role %s from %s' % (role, name))
            raise gen.Return()

        try:
            self.log.info('Adding role %s to %s' % (role, name))
            yield self.api_call(self.iam_conn.add_role_to_instance_profile,
                                name, role)
        except BotoServerError as e:
            if e.status != 409:
                raise exceptions.RecoverableActorFailure(
                    'An unexpected API error occurred: %s' % e)

    @gen.coroutine
    def _remove_role(self, name, role):
        """Removes a role assigned to an Instance Profile.

        args:
            name: The name of the InstanceProfile we're managing
            role: The name of the role to remove
        """
        if self._dry:
            self.log.warning('Would remove role %s from %s' % (role, name))
            raise gen.Return()

        try:
            self.log.info('Removing role %s from %s' % (role, name))
            yield self.api_call(
                self.iam_conn.remove_role_from_instance_profile,
                name, role)
        except BotoServerError as e:
            if e.status != 404:
                raise exceptions.RecoverableActorFailure(
                    'An unexpected API error occurred: %s' % e)

    @gen.coroutine
    def _ensure_role(self, name, role):
        """Ensures that an Instance Profile role is set correctly.

        Adds, Deletes or Changes the Role assigned to an Instance Profile.

        args:
            name: The IAM Instance Profile we're managing
            role: The desired role (or None)
        """
        existing = None
        try:
            raw = yield self.api_call(self.iam_conn.get_instance_profile, name)
            existing = (raw['get_instance_profile_response']
                           ['get_instance_profile_result']
                           ['instance_profile']
                           ['roles']
                           ['member']
                           ['role_name'])
        except BotoServerError as e:
            if e.status != 404:
                raise exceptions.RecoverableActorFailure(
                    'An unexpected API error occurred: %s' % e)
        except KeyError:
            # Profile is not a member of any roles
            pass

        if not existing and not role:
            raise gen.Return()
        elif existing and not role:
            try:
                yield self._remove_role(name, existing)
            except StopIteration:
                return
        elif not existing and role:
            yield self._add_role(name, role)
        elif existing != role:
            yield self._remove_role(name, existing)
            yield self._add_role(name, role)

    @gen.coroutine
    def _execute(self):
        name = self.option('name')
        state = self.option('state')
        role = self.option('role')

        yield self._ensure_entity(name, state)
        if state == 'absent':
            raise gen.Return()

        if role is not None:
            yield self._ensure_role(name, role)



"""Simple integration tests for the AWS CloudFormation actors."""

from nose.plugins.attrib import attr
import uuid
import logging

from tornado import testing

# from kingpin.actors import exceptions
from kingpin.actors.aws import cloudformation

__author__ = 'Matt Wise <matt@nextdoor.com>'

log = logging.getLogger(__name__)
logging.getLogger('boto').setLevel(logging.INFO)

UUID = uuid.uuid4().hex


class IntegrationCreate(testing.AsyncTestCase):

    """High Level CloudFormation Testing.

    These tests will check two things:
    * Create a super-simple CloudFormation stack
    * Delete that same stack

    Requirements:
        Your AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY must have access to
        create CF stacks. The stack we create is extremely simple, and should
        impact none of your AWS resources. The stack creates a simple S3
        bucket, so your credentials must have access to create that buckets.

    Note, these tests must be run in-order. The order is defined by
    their definition order in this file. Nose follows this order according
    to its documentation:

        http://nose.readthedocs.org/en/latest/writing_tests.html
    """

    integration = True

    region = 'us-east-1'
    bucket_name = 'kingpin-%s' % UUID

    @attr('aws', 'integration')
    @testing.gen_test(timeout=600)
    def integration_01_create_stack(self):
        actor = cloudformation.Create(
            'Create Stack',
            {'region': self.region,
             'name': self.bucket_name,
             'template':
                 'examples/test/aws.cloudformation/cf.integration.json',
             'parameters': {
                 'BucketName': self.bucket_name,
                 'BucketPassword': UUID,
                 'Metadata': UUID,
             }})

        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=60)
    def integration_02_create_duplicate_stack_should_fail(self):
        actor = cloudformation.Create(
            'Create Stack',
            {'region': self.region,
             'name': self.bucket_name,
             'template':
                 'examples/test/aws.cloudformation/cf.integration.json',
             'parameters': {
                 'BucketName': self.bucket_name,
                 'BucketPassword': UUID,
                 'Metadata': UUID,
             }})

        with self.assertRaises(cloudformation.StackAlreadyExists):
            yield actor.execute()

    @attr('aws', 'integration')
    @testing.gen_test(timeout=600)
    def integration_03_delete_stack(self):
        actor = cloudformation.Delete(
            'Delete Stack',
            {'region': self.region,
             'name': self.bucket_name})

        done = yield actor.execute()
        self.assertEqual(done, None)


class IntegrationStack(testing.AsyncTestCase):

    """High Level CloudFormation Stack Testing.

    These tests will check two things:
    * Create a super-simple CloudFormation stack
    * Update the stack
    * Delete that same stack

    Requirements:
        Your AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY must have access to
        create CF stacks. The stack we create is extremely simple, and should
        impact none of your AWS resources. The stack creates a simple S3
        bucket, so your credentials must have access to create that buckets.

    Note, these tests must be run in-order. The order is defined by
    their definition order in this file. Nose follows this order according
    to its documentation:

        http://nose.readthedocs.org/en/latest/writing_tests.html
    """

    integration = True

    region = 'us-east-1'
    bucket_name = 'kingpin-stack-%s' % UUID

    @attr('aws', 'integration')
    @testing.gen_test(timeout=600)
    def integration_01a_ensure_stack(self):
        actor = cloudformation.Stack(
            options={
                'region': self.region,
                'state': 'present',
                'name': self.bucket_name,
                'template':
                    'examples/test/aws.cloudformation/cf.integration.json',
                'parameters': {
                    'BucketName': self.bucket_name,
                    'BucketPassword': UUID,
                    'Metadata': UUID
                }})

        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=600)
    def integration_01b_ensure_stack_still_there(self):
        actor = cloudformation.Stack(
            options={
                'region': self.region,
                'state': 'present',
                'name': self.bucket_name,
                'template':
                    'examples/test/aws.cloudformation/cf.integration.json',
                'parameters': {
                    'BucketName': self.bucket_name,
                    'BucketPassword': UUID,
                    'Metadata': UUID
                }})

        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=600)
    def integration_02_changing_password_should_be_a_noop(self):
        #  This should pretty much do nothing.. if it did trigger a ChangeSet,
        #  we would actually fail because we're issuing a ChangeSet where no
        #  resources are actually modified. Thus, if this succeeds, we know
        #  that no stack change was made.
        actor = cloudformation.Stack(
            options={
                'region': self.region,
                'state': 'present',
                'name': self.bucket_name,
                'template':
                    'examples/test/aws.cloudformation/cf.integration.json',
                'parameters': {
                    'BucketName': self.bucket_name,
                    'BucketPassword': 'test',
                    'Metadata': UUID
                }})

        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=600)
    def integration_03_update_by_overriding_default(self):
        actor = cloudformation.Stack(
            options={
                'region': self.region,
                'state': 'present',
                'name': self.bucket_name,
                'template':
                    'examples/test/aws.cloudformation/cf.integration.json',
                'parameters': {
                    'BucketName': self.bucket_name,
                    'BucketPassword': UUID,
                    'DefaultParam': 'OverriddenValue',
                    'Metadata': UUID
                }})

        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=600)
    def integration_04a_update_bucket_name(self):
        actor = cloudformation.Stack(
            options={
                'region': self.region,
                'state': 'present',
                'name': self.bucket_name,
                'template':
                    'examples/test/aws.cloudformation/cf.integration.json',
                'parameters': {
                    'BucketName': '%s-updated' % self.bucket_name,
                    'BucketPassword': UUID,
                    'Metadata': UUID
                }})

        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=600)
    def integration_04b_update_bucket_name_second_time_should_work(self):
        actor = cloudformation.Stack(
            options={
                'region': self.region,
                'state': 'present',
                'name': self.bucket_name,
                'template':
                    'examples/test/aws.cloudformation/cf.integration.json',
                'parameters': {
                    'BucketName': '%s-updated' % self.bucket_name,
                    'BucketPassword': UUID,
                    'Metadata': UUID
                }})

        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=600)
    def integration_05a_delete_stack(self):
        actor = cloudformation.Stack(
            options={
                'region': self.region,
                'state': 'absent',
                'name': self.bucket_name,
                'template':
                    'examples/test/aws.cloudformation/cf.integration.json',
                'parameters': {
                    'BucketName': '%s-updated' % self.bucket_name,
                    'BucketPassword': UUID,
                    'Metadata': UUID
                }})

        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=600)
    def integration_05b_ensure_stack_absent(self):
        actor = cloudformation.Stack(
            options={
                'region': self.region,
                'state': 'absent',
                'name': self.bucket_name,
                'template':
                    'examples/test/aws.cloudformation/cf.integration.json',
                'parameters': {
                    'BucketName': '%s-updated' % self.bucket_name,
                    'BucketPassword': UUID,
                    'Metadata': UUID
                }})

        done = yield actor.execute()
        self.assertEqual(done, None)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.group`
^^^^^^^^^^^^^^^^^^^^^^^^^^^

Group a series of other `BaseActor` into either synchronous
or asynchronous stages.
"""

import logging

from tornado import gen
import demjson

from kingpin import utils as kp_utils
from kingpin.actors import base
from kingpin.actors import exceptions
from kingpin.actors import utils
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


class BaseGroupActor(base.BaseActor):

    """Group together a series of other `kingpin.actors.base.BaseActor` objects

    :acts:
      [ <list of `kingpin.actors.base.BaseActor` objects  to execute> ]

    """

    # By default, group actors have no timeout. We rely on the individual
    # actors to expire on their own. This is, of course, overrideable in the
    # JSON.
    default_timeout = None

    all_options = {
        'contexts': ((dict, str, list), [], "List of contextual hashes."),
        'acts': (list, REQUIRED, "Array of actor definitions.")
    }

    # Override the BaseActor strict_init_context setting. Since there may be
    # nested-groups that have their own context tokens, we do not require
    # that all of the {KEY}'s inside of the self._options dict are filled in
    # the moment that this actor is instantiated.
    strict_init_context = False

    # Do not remove remove escape sequence from escaped tokens. This will be
    # done later by another actor. Otherwise we risk remove the escapes and
    # failing because the token isn't found by a sub actor.
    remove_escape_sequence = False

    def __init__(self, *args, **kwargs):
        """Initializes all of the sub actors.

        By actually initializing all of the Actors supplied to us during the
        __init__, we effectively do a full instantiation of every Actor defined
        in the supplied JSON all at once and upfront long before we try to
        execute any code. This greatly increases our chances of catching JSON
        errors because every single object is pre-initialized before we ever
        begin executing any of our steps.

        *Note about init_tokens:*
          The group.BaseActor and misc.Macro actors support the concept of
          externally supplied data (usually os.environ) being used as available
          tokens for %TOKEN% parsing when reading JSON/YAML scripts. By passing
          this data between these three actors, we are able to allow nested
          token passing.

          See `Token-replacement <basicuse.html#token-replacement>` for more
          info.
        """
        super(BaseGroupActor, self).__init__(*args, **kwargs)

        # DEPRECATE IN v0.5.0
        if type(self.option('contexts')) == dict:
            try:
                filename = self.option('contexts').get('file', '')
                open(filename)
            except IOError as e:
                self.log.error('Option `contexts` must have valid `file`. '
                               'Received: %s' % filename)
                raise exceptions.InvalidOptions(e)
        # END DEPRECATION

        # Pre-initialize all of our actions!
        self._actions = self._build_actions()

    def get_orgchart(self, parent=''):
        """Generate an orgchart for all the `acts` specified."""

        ret = super(BaseGroupActor, self).get_orgchart(parent=parent)
        group_id = str(id(self))
        for act in self._actions:
            ret = ret + act.get_orgchart(parent=group_id)

        return ret

    def _build_actions(self):
        """Builds either a single set of actions, or multiple sets.

        If no 'contexts' were passed in, then we simply build the actors that
        are defined in the 'acts' option for the group.

        If any 'contexts' were passed in, then this method will create as many
        groups of actions as there are in the list of contexts. For each dict
        in the 'contexts' list, a new group of actors is created with that
        information.

        Note: Because groups may contain nested group actors, any options
        passed into this actors 'init_context' are also passed into the
        actors that we're intantiating.
        """
        contexts = self.option('contexts')
        if not contexts:
            return self._build_action_group(self._init_context)

        # If the data passed into the 'contexts' is a list of dicts, we take it
        # as is and do nothing to it.
        if type(contexts) == list:
            context_data = self.option('contexts')
        # DEPRECATE IN v0.5.0
        elif type(contexts) == dict:
            context_string = open(contexts['file']).read()
            context_string = kp_utils.populate_with_tokens(
                string=context_string,
                tokens=contexts.get('tokens', {}),
                strict=True)
            context_data = demjson.decode(context_string)
        # END DEPRECATION

        # If the data passed in is a string, it must be a pointer to a file
        # with contexts in it. We read that file, and we parse it for any
        # missing tokens. We use the "init tokens" that made it into this actor
        # as available token substitutions.
        elif isinstance(contexts, str):
            context_data = kp_utils.convert_script_to_dict(
                contexts, self._init_tokens)

        actions = []
        for context in context_data:
            combined_context = dict(list(self._init_context.items()) +
                                    list(context.items()))
            self.log.debug('Inherited context %s' %
                           list(self._init_context.items()))
            self.log.debug('Specified context %s' % list(context.items()))
            self.log.debug('Building acts with parameters: %s' %
                           combined_context)
            for action in self._build_action_group(context=combined_context):
                actions.append(action)

        return actions

    def _build_action_group(self, context=None):
        """Build up all of the actors we need to execute.

        Builds a list of actors to execute and returns the list. The list can
        then either be yielded as a whole (for an async operation), or
        individually (for a synchronous operation).

        Returns:
            A list of references to <actor objects>.
        """
        actions = []
        self.log.debug('Building %s actors' % len(self.option('acts')))
        for act in self.option('acts'):
            act['init_context'] = context.copy()
            act['init_tokens'] = self._init_tokens.copy()
            actor = utils.get_actor(act, dry=self._dry)
            actions.append(actor)
            self.log.debug('Actor %s built' % actor)
        return actions

    def _get_exc_type(self, exc_list):
        """Return Unrecoverable exception if at least one is in exc_list.

        Takes in a list of exceptions, and returns either a
        RecoverableActorFailure or an UnrecoverableActorFailure based on the
        exceptions that were passed in.

        Args:
            exc_list: List of Exception objects

        Returns:
            RecoverableActorFailure or UnrecoverableActorFailure
        """
        # Start by assuming we're going to be a RecoverableActorFailure
        wrapper_base = exceptions.RecoverableActorFailure
        for exc in exc_list:
            if isinstance(exc, exceptions.UnrecoverableActorFailure):
                wrapper_base = exceptions.UnrecoverableActorFailure
        return wrapper_base

    @gen.coroutine
    def _execute(self):
        """Executes the actions configured, and returns.

        Note: Expects the sub-class to implement self._run_actions()

        If an actor execution fails in _run_actions(), then that exception is
        raised up the stack.
        """
        self.log.info('Beginning %s actions' % len(self._actions))
        yield self._run_actions()
        raise gen.Return()


class Sync(BaseGroupActor):

    """Execute a series of `kingpin.actors.base.BaseActor` synchronously.

    Groups together a series of Actors and executes them synchronously
    in the order that they were defined.

    **Options**

    :acts:
      An array of individual Actor definitions.

    :contexts:

      This variable can be one of two formats:

      * A list of dictionaries with *contextual tokens* to pass into the actors
        at instantiation time. If the list has more than one element, then
        every actor defined in ``acts`` will be instantiated once for each item
        in the ``contexts`` list.
      * A string that points to a file with a list of contexts, just like the
        above dictionary.
      * (_Deprecation warning, this is going away in v0.4.0. Use the 'str'
        method above!_) A dictionary of ``file`` and ``tokens``. The file
        should be a relative path with data formatted same as stated above. The
        tokens need to be the same format as a Macro actor: a dictionary
        passing token data to be used.


    **Timeouts**

    Timeouts are disabled specifically in this actor. The sub-actors can still
    raise their own `kingpin.actors.exceptions.ActorTimedOut` exceptions, but
    since the group actors run an arbitrary number of sub actors, we have
    chosen to not have this actor specifically raise its own
    `kingpin.actors.exceptions.ActorTimedOut` exception unless the user sets
    the ``timeout`` setting.

    **Examples**

    Creates two arrays ... but sleeps 60 seconds between the two, then
    does not sleep at all after the last one:

    .. code-block:: json

       { "desc": "Clone, then sleep ... then clone, then sleep shorter...",
         "actor": "group.Sync",
         "options": {
           "contexts": [
             { "ARRAY": "First", "SLEEP": "60", },
             { "ARRAY": "Second", "SLEEP": "0", }
           ],
           "acts": [
             { "desc": "do something",
               "actor": "server_array.Clone",
               "options": {
                 "source": "template",
                 "dest": "{ARRAY}"
               }
             },
             { "desc": "sleep",
               "actor": "misc.Sleep",
               "options": {
                 "sleep": "{SLEEP}",
               }
             }
           ]
         }
       }

    Alternatively if no `contexts` are needed you can use the `array` syntax.

    .. code-block:: json

       [
         {
           "actor": "server_array.Clone",
           "options": {
             "source": "template",
             "dest": "%ARRAY%"
           }
         },
         {
           "actor": "misc.Sleep",
           "options": { "sleep": 30 }
         }
       ]

    **Dry Mode**

    Passes on the Dry mode setting to the acts that are called. Does **not**
    stop execution when one of the acts fails. Instead Group actor will finish
    all acts with warnings, and raise an error at the end of execution.

    This provides the user with an insight to all the errors that are possible
    to encounter, rather than abort and quit on the first one.

    **Failure**

    In the event that an act fails, this actor will return the failure
    immediately.  Because the acts are executed in-order of definition, the
    failure will prevent any further acts from executing.

    The behavior is different in the dry run (read above.)
    """

    @gen.coroutine
    def _run_actions(self):
        """Synchronously executes all of the Actor.execute() methods.

        If any one actor fails, we prevent execution of the rest of the actors.
        During a dry run - all acts are executed, and a warning is displayed.

        raises:
            In dry run - worst of all the raised errors.
            In real run - the first of the exceptions.
        """

        errors = []

        for act in self._actions:
            self.log.debug('Beginning "%s"..' % act._desc)
            try:
                yield act.execute()
            except exceptions.ActorException as e:
                if self._dry:
                    self.log.error('%s failed: %s' % (act._desc, str(e)))
                    self.log.warning('Continuing since this is a dry run.')
                    errors.append(e)
                else:
                    self.log.error('Aborting sequential execution because '
                                   '"%s" failed' % act._desc)
                    raise

        if errors:
            ExcType = self._get_exc_type(errors)
            raise ExcType('Exceptions raised by %s of %s actors in "%s".' % (
                          len(errors), len(self._actions), self._desc))


class Async(BaseGroupActor):

    """Execute several `kingpin.actors.base.BaseActor` objects asynchronously.

    Groups together a series of Actors and executes them asynchronously -
    waiting until all of them finish before returning.

    **Options**

    :concurrency:
      Max number of concurrent executions. This will fire off N executions
      in parallel, and continue with the remained as soon as the first
      execution is done. This is faster than creating N Sync executions.

    :acts:
      An array of individual Actor definitions.

    :contexts:

      This variable can be one of two formats:

      * A list of dictionaries with *contextual tokens* to pass into the actors
        at instantiation time. If the list has more than one element, then
        every actor defined in ``acts`` will be instantiated once for each item
        in the ``contexts`` list.
      * A dictionary of ``file`` and ``tokens``. The file should be a relative
        path with data formatted same as stated above. The tokens need to be
        the same format as a Macro actor: a dictionary passing token data to be
        used.

    **Timeouts**

    Timeouts are disabled specifically in this actor. The sub-actors can still
    raise their own `kingpin.actors.exceptions.ActorTimedOut` exceptions, but
    since the group actors run an arbitrary number of sub actors, we have
    chosen to not have this actor specifically raise its own
    `kingpin.actors.exceptions.ActorTimedOut` exception unless the user sets
    the ``timeout`` setting.

    **Examples**

    Clone two arrays quickly.

    .. code-block:: json

       { "desc": "Clone two arrays",
         "actor": "group.Async",
         "options": {
           "contexts": [
             { "ARRAY": "NewArray1" },
             { "ARRAY": "NewArray2" }
           ],
           "acts": [
             { "desc": "do something",
               "actor": "server_array.Clone",
               "options": {
                 "source": "template",
                 "dest": "{ARRAY}",
               }
             }
           ]
         }
       }

    **Dry Mode**

    Passes on the Dry mode setting to the sub-actors that are called.

    **Failure**

    In the event that one or more ``acts`` fail in this group, the entire group
    acts will return a failure to Kingpin. Because multiple actors are
    executing all at the same time, the all of these actors will be allowed to
    finish before the failure is returned.
    """

    all_options = {
        'concurrency': (int, 0, "Max number of concurrent executions."),
        'contexts': ((dict, str, list), [], "List of contextual hashes."),
        'acts': (list, REQUIRED, "Array of actor definitions.")
    }

    @gen.coroutine
    def _run_actions(self):
        """Asynchronously executes all of the Actor.execute() methods.

        All actors execute asynchronously, so we don't bother checking whether
        they've failed or not here. The BaseGroupActor will return a True/False
        based on whether or not all actors succeeded (True) or if one-or-more
        failed (False).
        """

        # This is an interesting tornado-ism. Here we generate and fire off
        # each of the acts asynchronously into the IOLoop, and we record
        # references to those tasks. However, we don't yield (wait) on them to
        # finish.
        tasks = []

        if self.option('concurrency'):
            self.log.info('Concurrency set to %s' % self.option('concurrency'))

        for act in self._actions:
            tasks.append(act.execute())

            if not self.option('concurrency'):
                # No concurrency limit - continue the loop without checks.
                continue

            running_tasks = len([t for t in tasks if not t.done()])

            if running_tasks < self.option('concurrency'):
                # We can queue more tasks, continue the loop to add one more.
                continue

            self.log.debug('Concurrency saturated. Waiting...')
            while running_tasks >= self.option('concurrency'):
                yield gen.moment
                running_tasks = len([t for t in tasks if not t.done()])

            self.log.debug('Concurrency desaturated: %s<%s. Continuing.' % (
                running_tasks, self.option('concurrency')))

        # Now that we've fired them off, we walk through them one-by-one and
        # check on their status. If they've raised an exception, we catch it
        # and log it into a list for further processing.
        errors = []
        for t in tasks:
            try:
                yield t
            except exceptions.ActorException as e:
                errors.append(e)

        # Now, if there are exceptions in the list, we generate the appropriate
        # exception type (recoverable vs unrecoverable), and raise it up the
        # stack. The individual exceptions are swallowed here, but thats OK
        # because the BaseActor for each of the acts that failed has already
        # handled printing out the log message with the failure.
        if errors:
            ExcType = self._get_exc_type(errors)
            raise ExcType('Exceptions raised by %s of %s actors in "%s".' % (
                          len(errors), len(self._actions), self._desc))



"""Tests for the actors.librato package"""

import json
import mock
import io

from tornado import gen
from tornado import httpclient
from tornado import testing

from kingpin.actors import exceptions
from kingpin.actors import librato


class FakeHTTPClientClass(object):
    '''Fake HTTPClient object for testing'''
    response_value = None

    @gen.coroutine
    def fetch(self, *args, **kwargs):
        raise gen.Return(self.response_value)


class FakeExceptionRaisingHTTPClientClass(object):
    '''Fake HTTPClient object for testing'''
    response_value = None

    @gen.coroutine
    def fetch(self, *args, **kwargs):
        raise self.response_value


class TestLibratoAnnotation(testing.AsyncTestCase):
    """Unit tests for the Librato Annotation actor."""

    def setUp(self, *args, **kwargs):
        # For most tests, mock out the TOKEN
        super(TestLibratoAnnotation, self).setUp()
        librato.TOKEN = 'Unittest'
        librato.EMAIL = 'Unittest'

    @testing.gen_test
    def test_init_without_token(self):
        # Un-set the token now and make sure the init fails
        librato.TOKEN = None
        with self.assertRaises(exceptions.InvalidCredentials):
            librato.Annotation(
                'Unit Test Action',
                {'title': 'unittest',
                 'description': 'unittest',
                 'name': 'unittest'})

    @testing.gen_test
    def test_init_without_email(self):
        # Un-set the token now and make sure the init fails
        librato.EMAIL = None
        with self.assertRaises(exceptions.InvalidCredentials):
            librato.Annotation(
                'Unit Test Action',
                {'title': 'unittest',
                 'description': 'unittest',
                 'name': 'unittest'})

    @testing.gen_test
    def test_execute_with_400(self):
        actor = librato.Annotation(
            'Unit Test Action',
            {'title': 'unittest',
             'description': 'unittest',
             'name': 'unittest'})

        http_response = httpclient.HTTPError(
            code=400, response={})

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeExceptionRaisingHTTPClientClass()
            m.return_value.response_value = http_response

            with self.assertRaises(exceptions.BadRequest):
                yield actor._execute()

    @testing.gen_test
    def test_execute_with_401(self):
        actor = librato.Annotation(
            'Unit Test Action',
            {'title': 'unittest',
             'description': 'unittest',
             'name': 'unittest'})

        http_response = httpclient.HTTPError(
            code=401, response={})

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeExceptionRaisingHTTPClientClass()
            m.return_value.response_value = http_response

            with self.assertRaises(exceptions.InvalidCredentials):
                yield actor._execute()

    @testing.gen_test
    def test_execute_with_unknown_exception(self):
        actor = librato.Annotation(
            'Unit Test Action',
            {'title': 'unittest',
             'description': 'unittest',
             'name': 'unittest'})

        http_response = httpclient.HTTPError(
            code=123, response={})

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeExceptionRaisingHTTPClientClass()
            m.return_value.response_value = http_response

            with self.assertRaises(httpclient.HTTPError):
                yield actor._execute()

    @testing.gen_test
    def test_execute(self):
        actor = librato.Annotation(
            'Unit Test Action',
            {'title': 'unittest',
             'description': 'unittest',
             'name': 'unittest'})

        response_dict = {'status': 'sent'}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPResponse(
            httpclient.HTTPRequest('/'), code=200,
            buffer=io.StringIO(response_body))

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeHTTPClientClass()
            m.return_value.response_value = http_response
            res = yield actor._execute()
            self.assertEqual(res, None)

    @testing.gen_test
    def test_execute_dry_mode_response(self):
        actor = librato.Annotation(
            'Unit Test Action',
            {'title': 'unittest',
             'description': 'unittest',
             'name': 'unittest'})
        actor._dry = True

        response_dict = {'status': 'sent'}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPResponse(
            httpclient.HTTPRequest('/'), code=200,
            buffer=io.StringIO(response_body))

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeHTTPClientClass()
            m.return_value.response_value = http_response
            res = yield actor._execute()
            self.assertEqual(res, None)
